from collections import defaultdict
from uuid import UUID

from django.core.exceptions import ValidationError
from django.db.models import Q
from drf_spectacular.utils import OpenApiExample, OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import NotFound, PermissionDenied
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.response import Response

from arkindex.documents.models import (
    Corpus,
    Element,
    Entity,
    EntityLink,
    EntityRole,
    Transcription,
    TranscriptionEntity,
)
from arkindex.documents.serializers.elements import ElementSlimSerializer
from arkindex.documents.serializers.entities import (
    BaseEntitySerializer,
    CreateEntityRoleErrorResponseSerializer,
    EntityCreateSerializer,
    EntityLinkCreateSerializer,
    EntityLinkSerializer,
    EntityRoleSerializer,
    EntitySerializer,
    TranscriptionEntityDetailsSerializer,
    TranscriptionEntitySerializer,
)
from arkindex.process.models import WorkerVersion
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, DeprecatedMixin
from arkindex.project.openapi import UUID_OR_FALSE
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.users.models import Role


@extend_schema(tags=['entities'])
@extend_schema_view(
    get=extend_schema(operation_id='ListCorpusRoles', description='List all roles of a corpus'),
    post=extend_schema(
        description='Create a new entity role',
        responses={
            200: EntityRoleSerializer,
            400: CreateEntityRoleErrorResponseSerializer
        },
        examples=[OpenApiExample(
            status_codes=['400'],
            response_only=True,
            name="role-exists",
            value={'id': "55cd009d-cd4b-4ec2-a475-b060f98f9138", 'corpus': ["Role already exists in this corpus"]},
            description="Role already exists."
        )]
    )
)
class CorpusRoles(CorpusACLMixin, ListCreateAPIView):
    """
    List all roles in a corpus
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = EntityRoleSerializer
    queryset = EntityRole.objects.none()

    def get_queryset(self):
        return EntityRole.objects \
            .filter(corpus=self.get_corpus(self.kwargs['pk'])) \
            .order_by('parent_name', 'child_name')

    def perform_create(self, serializer):
        data = self.request.data
        if EntityRole.objects.filter(
            parent_name=data['parent_name'],
            child_name=data['child_name'],
            parent_type=data['parent_type'],
            child_type=data['child_type'],
            corpus_id=self.request.parser_context['kwargs']['pk']
        ).exists():
            raise serializers.ValidationError({
                'corpus': ['Role already exists in this corpus'],
                'id': self.request.parser_context['kwargs']['pk']
            })
        super().perform_create(serializer)


@extend_schema(tags=['entities'])
@extend_schema_view(
    patch=extend_schema(description='Partially update an entity'),
    put=extend_schema(description='Update an entity'),
    delete=extend_schema(description='Delete an entity')
)
class EntityDetails(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get all information about an entity
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = EntitySerializer

    def get_queryset(self):
        return Entity.objects \
            .select_related('corpus') \
            .filter(corpus__in=Corpus.objects.readable(self.request.user)) \
            .prefetch_related(
                'parents__role',
                'children__role',
                'parents__child',
                'parents__parent',
                'children__parent',
                'children__child',
                'corpus'
            )

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if self.request.method not in permissions.SAFE_METHODS and not self.has_access(obj.corpus, Role.Contributor.value):
            raise PermissionDenied(detail='You do not have write access to this corpus.')


@extend_schema_view(get=extend_schema(operation_id='ListEntityElements', tags=['entities']))
class EntityElements(ListAPIView):
    """
    Get all elements that have a link with the entity
    """
    serializer_class = ElementSlimSerializer
    # For OpenAPI type discovery: an entity's ID is in the path
    queryset = Entity.objects.none()

    def get_queryset(self):
        pk = self.kwargs['pk']
        metadata_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                metadatas__entity_id=pk
            ) \
            .select_related('type', 'corpus') \
            .prefetch_related('metadatas__entity', 'image__server')
        transcription_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                transcriptions__transcription_entities__entity_id=pk
            ).select_related('type', 'corpus') \
            .prefetch_related('metadatas__entity', 'image__server')
        return metadata_elements.union(transcription_elements) \
            .order_by('name', 'type')


@extend_schema_view(post=extend_schema(
    operation_id='CreateEntity',
    responses={201: EntitySerializer},
    tags=['entities'],
))
class EntityCreate(CreateAPIView):
    """
    Create a new entity
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityCreateSerializer

    def perform_create(self, serializer):
        name = serializer.validated_data['name']
        type = serializer.validated_data['type']
        corpus = serializer.validated_data['corpus']
        metas = serializer.validated_data['metas'] if 'metas' in serializer.data else None
        worker_version = serializer.validated_data['worker_version']
        worker_run = serializer.validated_data['worker_run']
        return Entity.objects.create(
            name=name,
            type=type,
            corpus=corpus,
            metas=metas,
            worker_version=worker_version,
            worker_run=worker_run
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        entity = EntitySerializer(self.perform_create(serializer))
        headers = self.get_success_headers(serializer.data)
        return Response(entity.data, status=status.HTTP_201_CREATED, headers=headers)


@extend_schema_view(post=extend_schema(operation_id='CreateEntityLink', tags=['entities']))
class EntityLinkCreate(CreateAPIView):
    """
    Create a new link between two entities with a role
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityLinkCreateSerializer


@extend_schema_view(post=extend_schema(
    operation_id='CreateTranscriptionEntity',
    tags=['entities'],
    parameters=[
        OpenApiParameter(
            'id',
            type=UUID,
            location=OpenApiParameter.PATH,
            description='ID of the transcription to link an entity to.',
            required=True,
        )
    ]
))
class TranscriptionEntityCreate(CreateAPIView):
    """
    Link an existing Entity to a given transcription with its position.
    """
    permission_classes = (IsVerified, )
    serializer_class = TranscriptionEntitySerializer
    # For OpenAPI type discovery: a transcription's ID is in the path
    queryset = Transcription.objects.none()

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if not self.request:
            # OpenAPI cannot pass request context
            return context

        # We need to use the default database to avoid stale read
        # when assigning a newly created entity on on a newly created transcription
        tr = Transcription.objects.using('default').filter(
            id=self.kwargs.get('pk'),
            element__corpus__in=Corpus.objects.writable(self.request.user)
        ).first()
        return {**context, 'transcription': tr}

    def perform_create(self, serializer):
        data = serializer.validated_data
        try:
            TranscriptionEntity(**data).clean()
        except ValidationError as e:
            raise serializers.ValidationError(serializers.as_serializer_error(e))
        super().perform_create(serializer)


@extend_schema_view(
    get=extend_schema(
        operation_id='ListTranscriptionEntities',
        tags=['entities'],
        parameters=[
            OpenApiParameter(
                'worker_version',
                type=UUID_OR_FALSE,
                description='Only include transcription entities created by a specific worker version. '
                'If set to `False`, only include transcription entities created by humans.',
                required=False,
            ),
            OpenApiParameter(
                'entity_worker_version',
                type=UUID_OR_FALSE,
                description='Only include transcription entities whose entity was created by a specific worker version. '
                'If set to `False`, only include transcription entities whose entity was created by humans.',
                required=False,
            )
        ]
    )
)
class TranscriptionEntities(ListAPIView):
    """
    List existing entities linked to a specific transcription
    """
    serializer_class = TranscriptionEntityDetailsSerializer
    # For OpenAPI type discovery: a transcription's ID is in the path
    queryset = Transcription.objects.none()

    def parse_worker_version(self, value):
        if value.lower() in ('false', '0'):
            return None
        try:
            validated = UUID(value)
        except (TypeError, ValueError):
            raise serializers.ValidationError(['Invalid UUID.'])

        if not WorkerVersion.objects.filter(id=validated).exists():
            raise serializers.ValidationError(['This worker version does not exist.'])

        return validated

    def get_queryset(self):
        filters = {
            'transcription__element__corpus__in': Corpus.objects.readable(self.request.user),
            'transcription_id': self.kwargs['pk'],
        }

        errors = defaultdict()
        if 'worker_version' in self.request.query_params:
            try:
                worker_version_id = self.parse_worker_version(self.request.query_params['worker_version'])
                filters['worker_version_id'] = worker_version_id
            except serializers.ValidationError as e:
                errors['worker_version'] = e.detail
        if 'entity_worker_version' in self.request.query_params:
            try:
                entity_worker_version_id = self.parse_worker_version(self.request.query_params['entity_worker_version'])
                filters['entity__worker_version_id'] = entity_worker_version_id
            except serializers.ValidationError as e:
                errors['entity_worker_version'] = e.detail
        if errors:
            raise ValidationError(errors)

        return TranscriptionEntity.objects \
            .filter(**filters) \
            .order_by('offset') \
            .prefetch_related('entity')


@extend_schema_view(
    get=extend_schema(
        operation_id='ListElementEntities',
        tags=['entities'],
        deprecated=True,
        parameters=[
            OpenApiParameter(
                'worker_version',
                type=UUID,
                description='Only include entities created by a specific worker version',
                required=False,
            )
        ]
    )
)
class ElementEntities(DeprecatedMixin, RetrieveAPIView):
    """
    List all entities linked to an element's transcriptions and metadata
    """


@extend_schema_view(
    get=extend_schema(
        operation_id='ListCorpusEntities',
        tags=['entities'],
        parameters=[
            OpenApiParameter(
                'parent',
                type=UUID,
                description='Restrict entities to those linked to all transcriptions of an element '
                            'and all its descendants. Note that links to metadata are ignored.',
                required=False,
            ),
            OpenApiParameter(
                'name',
                type=str,
                description='Filter entities by part of their name (case-insensitive)',
                required=False,
            ),
        ]
    )
)
class CorpusEntities(CorpusACLMixin, ListAPIView):
    """
    List all entities in a corpus.
    """
    serializer_class = BaseEntitySerializer

    def get_queryset(self):
        corpus = self.get_corpus(self.kwargs['pk'])
        queryset = corpus.entities.order_by('name', 'id')

        if self.request.query_params.get('parent'):
            try:
                parent_id = UUID(self.request.query_params.get('parent'))
            except (TypeError, ValueError):
                raise serializers.ValidationError({'parent': ['Parent ID is not a valid UUID']})

            parent_query = corpus.elements.filter(id=parent_id)
            if not parent_query.exists():
                raise NotFound

            # We want to include entities that are either related to the selected element, or to its descendants
            # Element.objects.get_descending() does not include the parent element, so we add it manually using a union.
            # We remove the ordering from get_descending as it would add an extra column to the query and cause
            # the union to fail.
            # Filtering using Q(element_id=…) | Q(element_id__in=…) is 100 times slower than filtering using a union.

            element_ids = parent_query.values('id').union(Element.objects.get_descending(parent_id).order_by())
            queryset = queryset.filter(transcriptions__element_id__in=element_ids)

        if self.request.query_params.get('name'):
            queryset = queryset.filter(name__icontains=self.request.query_params['name'])

        return queryset


@extend_schema_view(get=extend_schema(operation_id='ListElementLinks', tags=['entities']))
class ElementLinks(CorpusACLMixin, ListAPIView):
    """
    List all links where parent and child are linked to the element.\n\n
    Requires a **guest** access to the element corpus
    """
    serializer_class = EntityLinkSerializer

    def get_queryset(self):
        try:
            element = Element.objects.select_related('corpus').only('id', 'corpus').get(id=self.kwargs['pk'])
        except Element.DoesNotExist:
            raise NotFound

        if not self.has_read_access(element.corpus):
            raise PermissionDenied(detail='You do not have access to this element.')

        # Load entities linked by transcriptions
        entities_tr = Entity.objects.filter(transcriptions__element_id=element.id).prefetch_related('transcriptions')

        # Load entities linked by metadatas
        entities_meta = Entity.objects.filter(metadatas__element_id=element.id).prefetch_related('metadatas')

        # Now load all links belonging to those entities
        # It's several times faster to combine the queries in the final one
        # than combining them at the lower level (entities is slower than entities_tr + entities_meta)
        # We need to support cross references between transcriptions & metadata entities
        return EntityLink.objects.filter(
            Q(parent__in=entities_tr, child__in=entities_tr)
            | Q(parent__in=entities_tr, child__in=entities_meta)
            | Q(parent__in=entities_meta, child__in=entities_tr)
            | Q(parent__in=entities_meta, child__in=entities_meta)
        ).select_related('role', 'child', 'parent').order_by('parent__name')
