from datetime import datetime, timedelta, timezone

from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import serializers, status
from rest_framework.exceptions import ValidationError
from rest_framework.generics import ListCreateAPIView, RetrieveAPIView
from rest_framework.response import Response

from arkindex.documents.models import Corpus, CorpusExport, CorpusExportState
from arkindex.documents.serializers.export import CorpusExportSerializer
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.permissions import IsVerified
from arkindex.users.models import Role


@extend_schema(tags=['exports'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListExports',
        description='List all exports on a corpus.',
    ),
    post=extend_schema(
        operation_id='StartExport',
        request=None,
        description='Start a corpus export job.',
    )
)
class CorpusExportAPIView(CorpusACLMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = CorpusExportSerializer
    queryset = CorpusExport.objects.none()

    def get_queryset(self):
        return CorpusExport \
            .objects \
            .filter(corpus=self.get_corpus(self.kwargs['pk'])) \
            .select_related('user') \
            .order_by('-created')

    def post(self, *args, **kwargs):
        corpus = self.get_corpus(self.kwargs['pk'], role=Role.Contributor)

        if corpus.exports.filter(state__in=(CorpusExportState.Created, CorpusExportState.Running)).exists():
            raise ValidationError('An export is already running for this corpus.')

        available_exports = corpus.exports.filter(
            state=CorpusExportState.Done,
            created__gte=datetime.now(timezone.utc) - timedelta(hours=6)
        )
        if available_exports.exists():
            raise ValidationError('An export has already been made for this corpus in the last 6 hours.')

        export = corpus.exports.create(user=self.request.user)
        export.start()
        return Response(CorpusExportSerializer(export).data, status=status.HTTP_201_CREATED)


class DownloadExport(RetrieveAPIView):
    """
    Download a corpus export.
    """
    queryset = CorpusExport.objects.none()
    permission_classes = (IsVerified, )

    def get_queryset(self):
        return CorpusExport.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
            state=CorpusExportState.Done
        ).only('id')

    @extend_schema(
        operation_id='DownloadExport',
        tags=['exports'],
        responses={302: serializers.Serializer},
    )
    def get(self, *args, **kwargs):
        return Response(status=status.HTTP_302_FOUND, headers={'Location': self.get_object().s3_url})
