from rest_framework import serializers

from arkindex.documents.models import EntityType, MetaType
from arkindex.project.serializer_fields import EnumField


class SolrDocumentSerializer(serializers.Serializer):
    """
    Serializes a Solr document
    """
    id = serializers.UUIDField()

    parent_id = serializers.UUIDField()
    parent_name = serializers.CharField()
    parent_type = serializers.CharField()

    element_id = serializers.UUIDField()
    element_text = serializers.CharField()
    element_type = serializers.CharField()
    element_worker = serializers.CharField(allow_null=True)
    element_image = serializers.URLField(allow_null=True)

    transcription_id = serializers.UUIDField(allow_null=True)
    transcription_text = serializers.CharField(allow_null=True)
    transcription_confidence = serializers.FloatField(min_value=0, max_value=1, required=False)
    transcription_worker = serializers.CharField(allow_null=True)

    classification_id = serializers.UUIDField(allow_null=True)
    classification_name = serializers.CharField(allow_null=True)
    classification_confidence = serializers.FloatField(min_value=0, max_value=1, required=False)
    classification_worker = serializers.CharField(allow_null=True)

    metadata_id = serializers.UUIDField(allow_null=True)
    metadata_name = serializers.CharField(allow_null=True)
    metadata_text = serializers.CharField(allow_null=True)
    metadata_type = EnumField(MetaType, allow_null=True)
    metadata_worker = serializers.CharField(allow_null=True)

    entity_id = serializers.UUIDField(allow_null=True)
    entity_text = serializers.CharField(allow_null=True)
    entity_type = EnumField(EntityType, allow_null=True)
    entity_worker = serializers.CharField(allow_null=True)


class SolrFacetsSerializer(serializers.Serializer):
    """
    Serializes Solr facets
    """
    element_type = serializers.DictField()
    element_worker = serializers.DictField()

    transcription_worker = serializers.DictField()

    classification_name = serializers.DictField()
    classification_worker = serializers.DictField()

    metadata_name = serializers.DictField()
    metadata_type = serializers.DictField()
    metadata_worker = serializers.DictField()

    entity_type = serializers.DictField()
    entity_worker = serializers.DictField()


class CorpusSearchResultSerializer(serializers.Serializer):
    """
    Serializes Corpus Search results
    """
    count = serializers.IntegerField()
    number = serializers.IntegerField()
    next = serializers.URLField(allow_null=True)
    previous = serializers.URLField(allow_null=True)
    results = SolrDocumentSerializer(many=True, allow_null=True)
    facets = SolrFacetsSerializer()


class CorpusSearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a Solr query
    """
    query = serializers.CharField()
    sources = serializers.MultipleChoiceField([
        ('element', 'element'),
        ('transcription', 'transcription'),
        ('metadata', 'metadata'),
        ('entity', 'entity')
    ], default={'element', 'transcription', 'metadata', 'entity'})
    only_facets = serializers.BooleanField(default=False)
    page = serializers.IntegerField(default=1, min_value=1)

    # Facets
    element_type = serializers.CharField(required=False)
    element_worker = serializers.CharField(required=False)

    transcription_worker = serializers.CharField(required=False)

    classification_name = serializers.CharField(required=False)
    classification_worker = serializers.CharField(required=False)

    metadata_name = serializers.CharField(required=False)
    metadata_type = serializers.CharField(required=False)
    metadata_worker = serializers.CharField(required=False)

    entity_type = serializers.CharField(required=False)
    entity_worker = serializers.CharField(required=False)
