import logging
from typing import Optional
from uuid import UUID

from django.conf import settings
from django.db.models import Q
from django_rq import job
from rq import Retry, get_current_job

from arkindex.documents.managers import ElementQuerySet
from arkindex.documents.models import (
    Classification,
    Corpus,
    Element,
    ElementPath,
    EntityLink,
    MetaData,
    Selection,
    Transcription,
    TranscriptionEntity,
)
from arkindex.process.models import Process, ProcessElement, WorkerActivity, WorkerRun

logger = logging.getLogger(__name__)


@job('high', timeout=settings.RQ_TIMEOUTS['corpus_delete'], retry=Retry(max=3))
def corpus_delete(corpus_id: str) -> None:
    # Note that this can be None when the task is run outside of a RQ worker (e.g. unit test)
    rq_job = get_current_job()
    corpus = Corpus.objects.get(id=corpus_id)
    logger.info(f'Deleting {corpus!r}')

    # When a corpus has a top level type, deleting all the ElementTypes will cause an IntegrityError.
    if corpus.top_level_type_id is not None:
        corpus.top_level_type_id = None
        corpus.save(update_fields=['top_level_type_id'])

    # Delete all related objects, bypassing RESTRICTs deletion rules
    # and Django's way of loading everything into memory before deleting.
    querysets = [
        corpus.allowed_metadatas.all(),
        corpus.worker_version_cache.all(),
        WorkerRun.objects.filter(process__corpus_id=corpus_id),
        # Using two separate queries is a lot faster than using a OR for large corpora
        ProcessElement.objects.filter(process__corpus_id=corpus_id),
        ProcessElement.objects.filter(element__corpus_id=corpus_id),
        # Process-DataFile M2M with implicit model
        Process.files.through.objects.filter(process__corpus_id=corpus_id),
        Process.files.through.objects.filter(datafile__corpus_id=corpus_id),
        # Worker activities
        WorkerActivity.objects.filter(Q(element__corpus_id=corpus_id) | Q(process__corpus_id=corpus_id)),
        corpus.processes.all(),
        corpus.files.all(),
        MetaData.objects.filter(element__corpus_id=corpus_id),
        EntityLink.objects.filter(role__corpus_id=corpus_id),
        corpus.roles.all(),
        TranscriptionEntity.objects.filter(entity__corpus_id=corpus_id),
        TranscriptionEntity.objects.filter(transcription__element__corpus_id=corpus_id),
        corpus.entities.all(),
        Classification.objects.filter(element__corpus_id=corpus_id),
        corpus.ml_classes.all(),
        Transcription.objects.filter(element__corpus_id=corpus_id),
        ElementPath.objects.filter(element__corpus_id=corpus_id),
        Selection.objects.filter(element__corpus_id=corpus_id),
        corpus.elements.all(),
        corpus.types.all(),
        corpus.memberships.all(),
        corpus.exports.all(),
        Corpus.objects.filter(id=corpus_id),
    ]

    for i, queryset in enumerate(querysets):
        if rq_job:
            rq_job.set_progress(i / len(querysets))
            rq_job.set_description(f'Deleting {queryset.model.__name__} on corpus {corpus.name}')
        deleted_count = queryset._raw_delete(using='default')
        logger.info(f'Deleted {deleted_count} {queryset.model.__name__}')

    logger.info(f'Deleted corpus {corpus_id}')


@job('high', timeout=settings.RQ_TIMEOUTS['element_trash'])
def element_trash(queryset: ElementQuerySet, delete_children: bool) -> None:
    queryset.trash(delete_children=delete_children)


@job('high', timeout=settings.RQ_TIMEOUTS['worker_results_delete'])
def selection_worker_results_delete(corpus_id: str, version_id: Optional[str] = None) -> None:
    """
    Delete all Worker Results produced by any WorkerVersion or a specific one
    on a user's selected elements, including their children.
    """
    rq_job = get_current_job()
    assert rq_job is not None, 'This task can only be run in a RQ job context.'
    assert rq_job.user_id is not None, 'This task requires a user ID to be defined on the RQ job.'

    queryset = Selection.objects.filter(user_id=rq_job.user_id, element__corpus_id=corpus_id)
    total = queryset.count()
    for i, element_id in enumerate(queryset.values_list('element_id', flat=True).iterator()):
        rq_job.set_progress(i / total)
        worker_results_delete(corpus_id=corpus_id, version_id=version_id, element_id=element_id)


@job('high', timeout=settings.RQ_TIMEOUTS['worker_results_delete'])
def worker_results_delete(
        corpus_id: str,
        version_id: Optional[str],
        element_id: Optional[str],
        include_children: bool = True) -> None:
    """
    Recursively delete all Worker Results produced by any WorkerVersion or a specific one
    on a whole corpus, under a specified parent element (parent element included), or on a single element.
    """
    elements = Element.objects.filter(corpus_id=corpus_id)
    classifications = Classification.objects.filter(element__corpus_id=corpus_id)
    transcriptions = Transcription.objects.filter(element__corpus_id=corpus_id)
    # TranscriptionEntities that will be removed due to the removal of transcriptions
    transcription_entities = TranscriptionEntity.objects.filter(transcription__element__corpus_id=corpus_id)
    # TranscriptionEntities from this worker
    worker_transcription_entities = TranscriptionEntity.objects.filter(
        Q(transcription__element__corpus_id=corpus_id)
        | Q(entity__corpus_id=corpus_id)
    )
    metadata = MetaData.objects.filter(element__corpus_id=corpus_id)
    worker_activities = WorkerActivity.objects.filter(element__corpus_id=corpus_id)

    # When a version ID is defined, filter by the exact version ID
    if version_id:
        elements = elements.filter(worker_version_id=version_id)
        classifications = classifications.filter(worker_version_id=version_id)
        transcriptions = transcriptions.filter(worker_version_id=version_id)
        transcription_entities = transcription_entities.filter(transcription__worker_version_id=version_id)
        worker_transcription_entities = worker_transcription_entities.filter(worker_version_id=version_id)
        metadata = metadata.filter(worker_version_id=version_id)
        worker_activities = worker_activities.filter(worker_version_id=version_id)
    # Otherwise, select everything that has any worker version ID.
    # We use worker_version_id != None and not worker_version_id__isnull=False,
    # because isnull would cause an unnecessary LEFT JOIN query.
    else:
        elements = elements.exclude(worker_version_id=None)
        classifications = classifications.exclude(worker_version_id=None)
        transcriptions = transcriptions.exclude(worker_version_id=None)
        transcription_entities = transcription_entities.exclude(transcription__worker_version_id=None)
        worker_transcription_entities = worker_transcription_entities.exclude(worker_version_id=None)
        metadata = metadata.exclude(worker_version_id=None)
        worker_activities = worker_activities.exclude(worker_version_id=None)

    if element_id:
        # include_children causes a deletion *only* on the element's descendants.
        # To also delete on the element itself, we will call this task synchronously with the same arguments and include_children=False.
        # This is used to avoid filtering by Q(id=element_id) | Q(paths__path__overlap=[element_id]) all at once,
        # which was found to be up to 30× slower than running two separate deletions.
        if include_children:
            elements = elements.filter(paths__path__overlap=[element_id])
            classifications = classifications.filter(element__paths__path__overlap=[element_id])
            transcriptions = transcriptions.filter(element__paths__path__overlap=[element_id])
            # The same filter is applied to both TranscriptionEntity querysets
            transcription_entities_filter = Q(transcription__element__paths__path__overlap=[element_id])
            transcription_entities = transcription_entities.filter(transcription_entities_filter)
            worker_transcription_entities = worker_transcription_entities.filter(transcription_entities_filter)
            metadata = metadata.filter(element__paths__path__overlap=[element_id])
            worker_activities = worker_activities.filter(element__paths__path__overlap=[element_id])
        else:
            elements = elements.filter(id=element_id)
            classifications = classifications.filter(element_id=element_id)
            transcriptions = transcriptions.filter(element_id=element_id)
            # The same filter is applied to both TranscriptionEntity querysets
            transcription_entities_filter = Q(transcription__element_id=element_id)
            transcription_entities = transcription_entities.filter(transcription_entities_filter)
            worker_transcription_entities = worker_transcription_entities.filter(transcription_entities_filter)
            metadata = metadata.filter(element_id=element_id)
            worker_activities = worker_activities.filter(element_id=element_id)

    elements.trash()
    classifications.delete()
    # Delete TranscriptionEntities before transcriptions so that we can delete transcriptions using a single DELETE query
    transcription_entities.delete()
    worker_transcription_entities.delete()
    transcriptions._raw_delete(using='default')
    metadata.delete()
    worker_activities.delete()

    if element_id and include_children:
        # Delete worker results on the parent element itself only after the deletion on its descendants has been done.
        # This deletion could cause the parent element to be deleted if it is considered to be a worker result itself,
        # which would remove the related ElementPaths and cause us to loose track of which child elements
        # we were supposed to delete worker results on.
        worker_results_delete(
            corpus_id=corpus_id,
            version_id=version_id,
            element_id=element_id,
            include_children=False,
        )


@job('high', timeout=settings.RQ_TIMEOUTS['move_element'])
def move_element(source: Element, destination: Element) -> None:
    parents = Element.objects.filter(id__in=source.paths.values('path__last')).only('id')
    for parent in parents:
        parent.remove_child(source)
    source.add_parent(destination)


@job('high', timeout=settings.RQ_TIMEOUTS['move_element'])
def move_selection(corpus_id: UUID, destination: Element) -> None:
    rq_job = get_current_job()
    assert rq_job is not None, 'This task can only be run in a RQ job context.'
    assert rq_job.user_id is not None, 'This task requires a user ID to be defined on the RQ job.'

    queryset = Element.objects.filter(selections__user_id=rq_job.user_id, corpus_id=corpus_id)
    total = queryset.count()
    for i, item in enumerate(queryset):
        rq_job.set_progress(i / total)
        move_element(source=item, destination=destination)


@job('high', timeout=settings.RQ_TIMEOUTS['move_element'])
def add_parent_selection(corpus_id: UUID, parent: Element) -> None:
    rq_job = get_current_job()
    assert rq_job is not None, 'This task can only be run in a RQ job context.'
    assert rq_job.user_id is not None, 'This task requires a user ID to be defined on the RQ job.'

    queryset = Element.objects.filter(selections__user_id=rq_job.user_id, corpus_id=corpus_id)
    total = queryset.count()
    for i, item in enumerate(queryset):
        rq_job.set_progress(i / total)
        item.add_parent(parent)
