from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, MLClass
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestBulkClassification(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_run = cls.worker_version.worker_runs.get()

    def create_classifications_data(self, classifications, parent=None):
        return {
            "parent": parent or str(self.page.id),
            "worker_version": str(self.worker_version.id),
            "classifications": classifications,
        }

    def test_requires_login(self):
        response = self.client.post(reverse('api:classification-bulk'), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_wrong_acl(self):
        self.client.force_login(self.user)
        private_page = self.private_corpus.elements.create(
            type=self.private_corpus.types.create(slug='page'),
        )
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data(
                    [{"class_name": 'dog', "confidence": 0.99}],
                    parent=str(private_page.id)
                )
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                'parent': ['Invalid pk "{}" - object does not exist.'.format(private_page.id)]
            }
        )

    def test_worker_version(self):
        """
        Classifications are created and linked to a worker version
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data([
                    {
                        "class_name": 'dog',
                        "confidence": 0.99,
                        "high_confidence": True
                    },
                    {
                        "class_name": 'cat',
                        "confidence": 0.42,
                    }
                ])
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_cl, second_cl = self.page.classifications.values_list('id', flat=True)
        self.assertEqual(response.json(), {
            'parent': str(self.page.id),
            'worker_version': str(self.worker_version.id),
            'worker_run_id': None,
            'classifications': [
                {
                    'id': str(first_cl),
                    'class_name': 'dog',
                    'confidence': 0.99,
                    'high_confidence': True,
                    'state': 'pending',
                },
                {
                    'id': str(second_cl),
                    'class_name': 'cat',
                    'confidence': 0.42,
                    'high_confidence': False,
                    'state': 'pending',
                },
            ]
        })

        self.assertCountEqual(
            list(self.page.classifications.values_list(
                'ml_class__name',
                'confidence',
                'high_confidence',
                'worker_version'
            )),
            [
                ('dog', 0.99, True, self.worker_version.id),
                ('cat', 0.42, False, self.worker_version.id)
            ],
        )

    def test_worker_version_xor_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            "class_name": 'cat',
                            "confidence": 0.42,
                        }
                    ],
                    "worker_run_id": str(self.worker_run.id),
                    "worker_version": str(self.worker_version.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'non_field_errors': ['Only one of `worker_version` and `worker_run_id` may be set.'],
        })

    def test_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            "class_name": 'dog',
                            "confidence": 0.99,
                            "high_confidence": True
                        },
                        {
                            "class_name": 'cat',
                            "confidence": 0.42,
                        }
                    ],
                    "worker_run_id": str(self.worker_run.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_cl, second_cl = self.page.classifications.values_list('id', flat=True)
        self.assertEqual(response.json(), {
            'parent': str(self.page.id),
            'worker_version': None,
            'worker_run_id': str(self.worker_run.id),
            'classifications': [
                {
                    'id': str(first_cl),
                    'class_name': 'dog',
                    'confidence': 0.99,
                    'high_confidence': True,
                    'state': 'pending',
                },
                {
                    'id': str(second_cl),
                    'class_name': 'cat',
                    'confidence': 0.42,
                    'high_confidence': False,
                    'state': 'pending',
                },
            ]
        })

        self.assertCountEqual(
            list(self.page.classifications.values_list(
                'ml_class__name',
                'confidence',
                'high_confidence',
                'worker_version_id',
                'worker_run_id',
            )),
            [
                ('dog', 0.99, True, self.worker_version.id, self.worker_run.id),
                ('cat', 0.42, False, self.worker_version.id, self.worker_run.id),
            ],
        )

    def test_worker_run_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            "class_name": 'dog',
                            "confidence": 0.99,
                            "high_confidence": True
                        },
                        {
                            "class_name": 'cat',
                            "confidence": 0.42,
                        }
                    ],
                    "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'worker_run_id': ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.'],
        })

    def test_create_ml_class(self):
        """
        Adding classifications with non existing classes should automatically create them
        """
        self.assertEqual(MLClass.objects.filter(name="dog", corpus=self.corpus).count(), 0)
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data(
                    [{"class_name": 'dog', "confidence": 0.99}],
                )
            )
        self.assertEqual(MLClass.objects.filter(name="dog", corpus=self.corpus).count(), 1)

    def test_existing_ml_class(self):
        MLClass.objects.create(name="cat", corpus=self.corpus)
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data(
                    [{"class_name": 'cat', "confidence": 0.42}],
                )
            )
        self.assertEqual(MLClass.objects.filter(name="cat", corpus=self.corpus).count(), 1)

    def test_delete_worker_version(self):
        """
        Test the bulk classification API deletes previous classifications with a similar worker version
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data([
                    {
                        "class_name": 'dog',
                        "confidence": 0.99,
                        "high_confidence": True
                    },
                    {
                        "class_name": 'cat',
                        "confidence": 0.42,
                    }
                ])
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data([
                    {
                        "class_name": 'doggo',
                        "confidence": 0.5,
                    },
                    {
                        "class_name": 'catte',
                        "confidence": 0.85,
                        "high_confidence": True
                    }
                ])
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(
            list(self.page.classifications.values_list('ml_class__name', 'confidence', 'high_confidence')),
            [
                ('doggo', 0.5, False),
                ('catte', 0.85, True),
            ],
        )

    def test_delete_worker_run(self):
        """
        Test the bulk classification API deletes previous classifications with the same worker run
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            "class_name": 'dog',
                            "confidence": 0.99,
                            "high_confidence": True
                        },
                        {
                            "class_name": 'cat',
                            "confidence": 0.42,
                        }
                    ],
                    "worker_run_id": str(self.worker_run.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifications": [
                        {
                            "class_name": 'doggo',
                            "confidence": 0.5,
                        },
                        {
                            "class_name": 'catte',
                            "confidence": 0.85,
                            "high_confidence": True
                        }
                    ],
                    "worker_run_id": str(self.worker_run.id),
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertCountEqual(
            list(self.page.classifications.values_list('ml_class__name', 'confidence', 'high_confidence')),
            [
                ('doggo', 0.5, False),
                ('catte', 0.85, True),
            ],
        )

    def test_no_duplicates(self):
        """
        Test the bulk classification API prevents creating classifications with duplicate ML classes
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data([
                    {"class_name": 'dog', "confidence": 0.99},
                    {"class_name": 'dog', "confidence": 0.99},
                ])
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'classifications': ['Duplicated ML classes are not allowed from the same worker version or worker run.']
        })
