from uuid import uuid4

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Element, ElementPath
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestBulkElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_run = cls.worker_version.worker_runs.get()
        cls.payload = {
            'worker_version': str(cls.worker_version.id),
            'elements': [
                {
                    'name': 'A',
                    'type': 'act',
                    'polygon': [
                        [0, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    'name': 'B',
                    'type': 'surface',
                    # Reuses the same zone
                    'polygon': [
                        [0, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    'name': 'C',
                    'type': 'surface',
                    'polygon': [
                        [0, 0],
                        [0, 9],
                        [9, 9],
                        [9, 0],
                        [0, 0],
                    ]
                }
            ]
        }
        cls.payload_2 = {
            "worker_run_id": str(cls.worker_run.id),
            "elements": cls.payload["elements"]
        }

    def setUp(self):
        super().setUp()
        self.element = self.corpus.elements.get(name='Volume 2, page 1v')
        self.rotated_page = Element.objects.create(corpus=self.corpus, type=self.element.type, image=self.element.image, polygon=self.element.polygon, name='rotated_page', rotation_angle=90)
        self.mirrored_page = Element.objects.create(corpus=self.corpus, type=self.element.type, image=self.element.image, polygon=self.element.polygon, name='mirrored_page', mirrored=True)

    def test_bulk_create_requires_login(self):
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data=self.payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_bulk_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data=self.payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_bulk_create_writable_corpus(self):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data=self.payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_bulk_create_requires_parent_zone(self):
        self.element.image = None
        self.element.polygon = None
        self.element.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data={'elements': []},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element': ['Cannot create child elements on an element without a zone.']
        })

    def test_bulk_create_requires_elements(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data={'elements': []},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'elements': {'non_field_errors': ['This list may not be empty.']}})

    def test_bulk_create_missing_types(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                data={'elements': [
                    {
                        'name': 'nope',
                        'type': 'nope',
                        'polygon': [[0, 0], [0, 42], [13, 37]],
                    }
                ]},
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'non_field_errors': ["Element types with slugs nope do not exist in the parent element's corpus"]
        })

    def test_bulk_create(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                data=self.payload,
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        element_path = self.element.paths.get()
        a, b, c = Element \
            .objects \
            .get_descending(self.element.id) \
            .filter(type__slug__in=('act', 'surface')) \
            .order_by('name')
        a_path, b_path, c_path = a.paths.get(), b.paths.get(), c.paths.get()

        self.assertListEqual(
            response.json(),
            [
                {'id': str(a.id)},
                {'id': str(b.id)},
                {'id': str(c.id)}
            ]
        )

        self.assertEqual(a.name, 'A')
        self.assertEqual(b.name, 'B')
        self.assertEqual(c.name, 'C')
        self.assertEqual(a.type.slug, 'act')
        self.assertEqual(b.type.slug, 'surface')
        self.assertEqual(c.type.slug, 'surface')
        self.assertEqual(a.worker_version, self.worker_version)
        self.assertEqual(b.worker_version, self.worker_version)
        self.assertEqual(c.worker_version, self.worker_version)
        self.assertEqual(a.worker_run, None)
        self.assertEqual(b.worker_run, None)
        self.assertEqual(c.worker_run, None)
        self.assertEqual(a.image_id, self.element.image_id)
        self.assertEqual(b.image_id, self.element.image_id)
        self.assertEqual(c.image_id, self.element.image_id)
        self.assertTupleEqual(a.polygon.coords, ((0, 0), (0, 1), (1, 1), (1, 0), (0, 0)))
        self.assertTupleEqual(b.polygon.coords, ((0, 0), (0, 1), (1, 1), (1, 0), (0, 0)))
        self.assertTupleEqual(c.polygon.coords, ((0, 0), (0, 9), (9, 9), (9, 0), (0, 0)))
        self.assertEqual(a.rotation_angle, self.element.rotation_angle)
        self.assertEqual(b.rotation_angle, self.element.rotation_angle)
        self.assertEqual(c.rotation_angle, self.element.rotation_angle)
        self.assertEqual(a.mirrored, self.element.mirrored)
        self.assertEqual(b.mirrored, self.element.mirrored)
        self.assertEqual(c.mirrored, self.element.mirrored)

        self.assertListEqual(a_path.path, element_path.path + [self.element.id])
        self.assertListEqual(b_path.path, element_path.path + [self.element.id])
        self.assertListEqual(c_path.path, element_path.path + [self.element.id])
        self.assertEqual(a_path.ordering, 0)
        self.assertEqual(b_path.ordering, 0)
        self.assertEqual(c_path.ordering, 1)

    def test_bulk_create_multiple_parent_paths(self):
        parent = self.corpus.elements.create(
            name='Parent 2',
            type=self.corpus.types.first(),
        )
        self.element.add_parent(parent)
        element_path1, element_path2 = self.element.paths.order_by('path__0')

        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                data=self.payload,
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        a, b, c = Element \
            .objects \
            .get_descending(self.element.id) \
            .filter(type__slug__in=('act', 'surface')) \
            .order_by('name')

        a_path1, a_path2 = a.paths.order_by('path__0')
        b_path1, b_path2 = b.paths.order_by('path__0')
        c_path1, c_path2 = c.paths.order_by('path__0')

        self.assertListEqual(a_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(a_path2.path, element_path2.path + [self.element.id])
        self.assertListEqual(b_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(b_path2.path, element_path2.path + [self.element.id])
        self.assertListEqual(c_path1.path, element_path1.path + [self.element.id])
        self.assertListEqual(c_path2.path, element_path2.path + [self.element.id])

        self.assertEqual(a_path1.ordering, 0)
        self.assertEqual(a_path2.ordering, 0)
        self.assertEqual(b_path1.ordering, 0)
        self.assertEqual(b_path2.ordering, 0)
        self.assertEqual(c_path1.ordering, 1)
        self.assertEqual(c_path2.ordering, 1)

    def test_bulk_create_top_level(self):

        # Create a top level page, without any path
        element = self.corpus.elements.create(
            name="Top level page",
            type=self.corpus.types.get(slug='page'),
            image=self.element.image,
            polygon=self.element.polygon,
        )
        self.assertFalse(element.paths.exists())

        # Try to create a sub element on that zone
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(element.id)}),
                data={
                    'worker_version': str(self.worker_version.id),
                    'elements': [
                        {
                            'name': f'surface {i}',
                            'type': 'surface',
                            'polygon': [
                                [i * 100, i * 100],
                                [(i + 1) * 100, i * 100],
                                [(i + 1) * 100, (i + 1) * 100],
                                [i * 100, (i + 1) * 100],
                            ]
                        }
                        for i in range(1, 5)
                    ]
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # Check the created elements are children of element
        data = response.json()
        self.assertEqual(len(data), 4)
        paths = ElementPath.objects.filter(element_id__in=[elt['id'] for elt in data], path=[element.id])
        self.assertEqual(paths.count(), 4)

    def test_bulk_create_negative_polygon(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                {
                    'worker_version': str(self.worker_version.id),
                    'elements': [
                        {
                            'name': 'Blah',
                            'type': 'surface',
                            'polygon': [[0, 10], [10, 20], [30, 40], [50, 60], [0, -10]]
                        }
                    ]
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'elements': [
                {
                    'polygon': {
                        '4': {
                            '1': ['Ensure this value is greater than or equal to 0.']
                        }
                    }
                }
            ]
        })

    def test_bulk_create_inherit_rotation(self):
        """
        Created children elements inherit their parent's rotation angle
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.rotated_page.id)}),
            data=self.payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_elements = Element.objects.get_descending(self.rotated_page.id)
        for elt in created_elements:
            self.assertEqual(elt.rotation_angle, 90)

    def test_bulk_create_inherit_mirrored(self):
        """
        Created children elements inherit their parent's mirroring
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.mirrored_page.id)}),
            data=self.payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_elements = Element.objects.get_descending(self.mirrored_page.id)
        for elt in created_elements:
            self.assertEqual(elt.mirrored, True)

    def test_bulk_create_outside_image(self):
        """
        Cannot create elements outside their image
        """
        self.maxDiff = None
        self.client.force_login(self.user)
        payload = {
            'worker_version': str(self.worker_version.id),
            'elements': [
                {
                    'name': 'A',
                    'type': 'act',
                    'polygon': [
                        [0, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    'name': 'B',
                    'type': 'surface',
                    # Out of bounds
                    'polygon': [
                        [1003, 0],
                        [0, 1],
                        [1, 1],
                        [1, 0],
                        [0, 0],
                    ]
                },
                {
                    'name': 'C',
                    'type': 'surface',
                    # Also out of bounds
                    'polygon': [
                        [0, 0],
                        [0, 9],
                        [9, 1200],
                        [9, 0],
                        [0, 0],
                    ]
                }
            ]
        }
        response = self.client.post(
            reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
            data=payload,
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'elements': {
                '1': {'polygon': ["An element's polygon must not exceed its image's bounds."]},
                '2': {'polygon': ["An element's polygon must not exceed its image's bounds."]}
            }
        })

    def test_bulk_create_unknown_worker_run(self):
        self.client.force_login(self.user)
        random_uuid = str(uuid4())
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                {
                    'worker_run_id': random_uuid,
                    'elements': [
                        {
                            'name': 'Blah',
                            'type': 'surface',
                            'polygon': [[0, 10], [10, 20], [30, 40], [50, 60], [0, 10]]
                        }
                    ]
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_run_id': [f'Invalid pk "{random_uuid}" - object does not exist.']
        })

    def test_bulk_create_invalid_parameters_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                {
                    'worker_version': str(self.worker_version.id),
                    'worker_run_id': str(self.worker_run.id),
                    'elements': [
                        {
                            'name': 'Blah',
                            'type': 'surface',
                            'polygon': [[0, 10], [10, 20], [30, 40], [50, 60], [0, 10]]
                        }
                    ]
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['Only one of `worker_version` and `worker_run_id` may be set.']
        })

    def test_bulk_create_with_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse('api:elements-bulk-create', kwargs={'pk': str(self.element.id)}),
                data=self.payload_2,
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        element_path = self.element.paths.get()
        a, b, c = Element \
            .objects \
            .get_descending(self.element.id) \
            .filter(type__slug__in=('act', 'surface')) \
            .order_by('name')
        a_path, b_path, c_path = a.paths.get(), b.paths.get(), c.paths.get()

        self.assertListEqual(
            response.json(),
            [
                {'id': str(a.id)},
                {'id': str(b.id)},
                {'id': str(c.id)}
            ]
        )

        self.assertEqual(a.name, 'A')
        self.assertEqual(b.name, 'B')
        self.assertEqual(c.name, 'C')
        self.assertEqual(a.type.slug, 'act')
        self.assertEqual(b.type.slug, 'surface')
        self.assertEqual(c.type.slug, 'surface')
        self.assertEqual(a.worker_version, self.worker_run.version)
        self.assertEqual(b.worker_version, self.worker_run.version)
        self.assertEqual(c.worker_version, self.worker_run.version)
        self.assertEqual(a.worker_run, self.worker_run)
        self.assertEqual(b.worker_run, self.worker_run)
        self.assertEqual(c.worker_run, self.worker_run)
        self.assertEqual(a.image_id, self.element.image_id)
        self.assertEqual(b.image_id, self.element.image_id)
        self.assertEqual(c.image_id, self.element.image_id)
        self.assertTupleEqual(a.polygon.coords, ((0, 0), (0, 1), (1, 1), (1, 0), (0, 0)))
        self.assertTupleEqual(b.polygon.coords, ((0, 0), (0, 1), (1, 1), (1, 0), (0, 0)))
        self.assertTupleEqual(c.polygon.coords, ((0, 0), (0, 9), (9, 9), (9, 0), (0, 0)))
        self.assertEqual(a.rotation_angle, self.element.rotation_angle)
        self.assertEqual(b.rotation_angle, self.element.rotation_angle)
        self.assertEqual(c.rotation_angle, self.element.rotation_angle)
        self.assertEqual(a.mirrored, self.element.mirrored)
        self.assertEqual(b.mirrored, self.element.mirrored)
        self.assertEqual(c.mirrored, self.element.mirrored)

        self.assertListEqual(a_path.path, element_path.path + [self.element.id])
        self.assertListEqual(b_path.path, element_path.path + [self.element.id])
        self.assertListEqual(c_path.path, element_path.path + [self.element.id])
        self.assertEqual(a_path.ordering, 0)
        self.assertEqual(b_path.ordering, 0)
        self.assertEqual(c_path.ordering, 1)
