from django.core.management.base import BaseCommand, CommandError

from arkindex.documents.models import ElementType
from arkindex.process.models import Process, ProcessMode
from arkindex.project.argparse import CorpusArgument, ElementArgument, ProcessArgument, UserArgument


class Command(BaseCommand):
    help = 'Apply a template to build a set of Processes'

    def add_arguments(self, parser):
        parser.add_argument(
            'process',
            help='Process to use as a template for workers',
            type=ProcessArgument()
        )
        parser.add_argument(
            '--corpus',
            required=True,
            help='Corpus ID or name to create Processes on',
            type=CorpusArgument()
        )
        parser.add_argument(
            '--element',
            dest='elements',
            nargs='+',
            required=True,
            help='Top level element ids to build workflows: each element will get its process',
            type=ElementArgument(),
        )
        parser.add_argument(
            '--children-type',
            type=str,
            required=True,
            help='Element type slug to use to build the new process',
        )
        parser.add_argument(
            '--chunks',
            type=int,
            default=1,
            help='Number of chunks to build for the process',
        )
        parser.add_argument(
            '--creator',
            required=True,
            type=UserArgument(),
            help='Creator of the new process',
        )

    def handle(self, process, corpus, elements, children_type, *args, **options):
        # Check template (must be in worker mode and have some workers)
        if process.mode != ProcessMode.Workers:
            raise CommandError("Only workers processes are supported")
        if not process.worker_runs.exists():
            raise CommandError("This process has no worker runs defined")

        # Check new process related models (corpus, elements, type)
        for element in elements:
            if element.corpus != corpus:
                raise CommandError(f"Element {element.id} is not in corpus {corpus}")

        try:
            element_type = corpus.types.get(slug=children_type)
        except ElementType.DoesNotExist:
            raise CommandError(f"Element type {children_type} is not in corpus {corpus}")

        # Now build a clone for each top level element
        for element in elements:
            self.clone(process, element, element_type, options['creator'], options['chunks'])

    def clone(self, source, element, element_type, creator, chunks):
        """
        Clone a process configuration, on new elements
        """

        # Name is using the source's name, along with the element's name
        name = f"Run {source.name} on {element.name}" if source.name else element.name

        # Build a process that will load all specified children
        # elements from that top level element
        p = Process.objects.create(
            mode=ProcessMode.Workers,
            name=name,
            corpus=element.corpus,
            creator=creator,
            element=element,
            load_children=True,
            element_type=element_type,
        )

        print(f'Created Process {p.id}')

        # Build linear worker runs
        runs = {}
        for wr in source.worker_runs.all():
            runs[wr.id] = p.worker_runs.create(
                version=wr.version,
                parents=[]
            )

        # Set links
        for wr in source.worker_runs.filter(parents__len__gt=0):
            runs[wr.id].parents = [
                runs[parent].id
                for parent in wr.parents
            ]
            runs[wr.id].save()

        # Build and start workflow
        p.start(chunks=chunks)
        print(f'Started Process {p.id}')
