import urllib
from collections import defaultdict
from enum import Enum

from django.db.models import Q
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.process.models import (
    Process,
    Repository,
    Revision,
    Worker,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerType,
    WorkerVersion,
    WorkerVersionGPUUsage,
    WorkerVersionState,
)
from arkindex.process.serializers.git import RevisionWithRefsSerializer
from arkindex.process.utils import hash_object
from arkindex.project.serializer_fields import EnumField


class WorkerLightSerializer(serializers.ModelSerializer):
    """
    Serialize a simple repository worker
    """
    # We want to only have to specify the type as a slug (char) in stead of the actual workerType id
    type = serializers.SlugField()

    class Meta:
        model = Worker
        fields = ('id', 'name', 'type', 'slug')

    def validate_type(self, type):
        worker_type, _ = WorkerType.objects.get_or_create(slug=type, defaults={'display_name': type.capitalize()})
        return worker_type


class WorkerSerializer(WorkerLightSerializer):
    """
    Serialize a worker with its repository and slug
    """
    repository_id = serializers.UUIDField(required=True)

    class Meta(WorkerLightSerializer.Meta):
        model = Worker
        fields = WorkerLightSerializer.Meta.fields + ('repository_id', )


class WorkerTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = WorkerType
        fields = (
            'id',
            'slug',
            'display_name'
        )


class UserConfigurationFieldType(Enum):
    Int = 'int'
    Float = 'float'
    String = 'string'
    Enum = 'enum'
    Boolean = 'bool'
    Dict = 'dict'
    List = 'list'


class UserConfigurationFieldSerializer(serializers.Serializer):
    title = serializers.CharField()
    type = EnumField(UserConfigurationFieldType)
    subtype = EnumField(UserConfigurationFieldType, required=False)
    required = serializers.BooleanField(default=False)
    choices = serializers.ListField(required=False, allow_empty=False, allow_null=True)

    def to_internal_value(self, data):
        errors = defaultdict(list)

        allowed_fields = ['title', 'type', 'required', 'default', 'choices', 'subtype']
        data_types = {
            UserConfigurationFieldType.Int: [serializers.IntegerField(), int],
            UserConfigurationFieldType.Float: [serializers.FloatField(), float],
            UserConfigurationFieldType.String: [serializers.CharField(), str],
            UserConfigurationFieldType.Enum: [serializers.ChoiceField(choices=[]), None],
            UserConfigurationFieldType.Boolean: [serializers.BooleanField(), bool],
            UserConfigurationFieldType.Dict: [serializers.DictField(child=serializers.CharField()), dict],
            UserConfigurationFieldType.List: [serializers.ListField(), list]
        }

        for field in data:
            if field not in allowed_fields:
                errors[field].append(
                    'Configurable properties can only be defined using the following keys: title, type, required, default, subtype, choices.'
                )

        default_value = data.get('default')
        data = super().to_internal_value(data)

        field_type = data.get('type')
        choices = data.get('choices')
        subtype = data.get('subtype')

        if field_type == UserConfigurationFieldType.List and not subtype:
            errors['subtype'].append('The "subtype" field must be set for "list" type properties.')
        # Handle subtypes
        if subtype is not None:
            if field_type != UserConfigurationFieldType.List:
                errors['subtype'].append('The "subtype" field can only be set for a "list" type property.')
            if subtype not in [UserConfigurationFieldType.Int, UserConfigurationFieldType.Float, UserConfigurationFieldType.String, UserConfigurationFieldType.Boolean]:
                errors['subtype'].append('Subtype can only be int, float, bool or string.')
        # Handle enums
        if choices is not None:
            if field_type != UserConfigurationFieldType.Enum:
                errors['choices'].append('The "choices" field can only be set for an "enum" type property.')
            # If the configuration parameter is of enum type, an eventual default value won't match the field type
            if default_value and default_value not in choices:
                errors['default'].append(f'{default_value} is not an available choice.')
        # Handle everything else
        if default_value is not None and field_type != UserConfigurationFieldType.Enum:
            try:
                data_type, data_class = data_types[field_type]
                if not isinstance(default_value, data_class):
                    raise ValidationError
                data_type.to_internal_value(default_value)
                # For lists, check that list elements are of given subtype
                if field_type == UserConfigurationFieldType.List and not errors.get('subtype'):
                    _, data_subclass = data_types[subtype]
                    if any(not isinstance(item, data_subclass) for item in default_value):
                        errors['default'].append(f'All items in the default value must be of type {data_subclass.__name__}.')
            except ValidationError:
                errors['default'].append(f'Default value is not of type {field_type.value}.')
            except KeyError:
                errors['default'].append(f'Cannot check type: {field_type.value}.')

        if errors:
            raise ValidationError(errors)
        return data


class WorkerVersionSerializer(serializers.ModelSerializer):
    """
    Serialize a worker version
    """
    # State defaults to created when instantiating a WorkerVersion
    state = EnumField(WorkerVersionState, required=False)
    worker = WorkerLightSerializer(read_only=True)
    revision = serializers.UUIDField()
    gpu_usage = EnumField(WorkerVersionGPUUsage, required=False, default=WorkerVersionGPUUsage.Disabled)
    model_usage = serializers.BooleanField(required=False, default=False)

    # Serialize worker with its basic information
    class Meta:
        model = WorkerVersion
        fields = (
            'id',
            'configuration',
            'revision',
            'docker_image',
            'docker_image_iid',
            'docker_image_name',
            'state',
            'gpu_usage',
            'model_usage',
            'worker'
        )
        read_only_fields = ('docker_image_name',)
        # Avoid loading all revisions and all Ponos artifacts when opening this endpoint in a browser
        extra_kwargs = {
            'revision': {'style': {'base_template': 'input.html'}},
            'docker_image': {'style': {'base_template': 'input.html'}},
        }

    def to_representation(self, instance):
        self.fields['revision'] = RevisionWithRefsSerializer(read_only=True)
        return super(WorkerVersionSerializer, self).to_representation(instance)

    def validate_revision(self, revision_id):
        # Retrieve a revision from its ID without listing them with a PrimaryKeyRelatedField
        try:
            return Revision.objects.get(id=revision_id)
        except Revision.DoesNotExist:
            raise ValidationError({'revision': ['Revision with this ID does not exist.']})

    def validate_configuration(self, configuration):
        errors = defaultdict(list)
        user_configuration = configuration.get('user_configuration')
        if not user_configuration:
            return configuration
        field = serializers.DictField(child=UserConfigurationFieldSerializer())
        try:
            field.to_internal_value(user_configuration)
        except ValidationError as e:
            errors['user_configuration'].append(e.detail)
        if errors:
            raise ValidationError(errors)
        return configuration

    def validate(self, data):
        # Assert that a version set to available has a docker image
        state = data.get('state') or self.instance and self.instance.state
        if not state:
            return data

        # Determine the future docker image fields value considering those fields are nullable
        image = (
            data['docker_image'] if 'docker_image' in data
            else self.instance and self.instance.docker_image_id
        )
        image_iid = (
            data['docker_image_iid'] if 'docker_image_iid' in data
            else self.instance and self.instance.docker_image_iid
        )

        if state == WorkerVersionState.Available and (not image or not image_iid):
            raise ValidationError([
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ])

        return data


class WorkerVersionEditSerializer(WorkerVersionSerializer):
    """
    Serialize a worker version run with a non editable revision field
    """
    revision = serializers.UUIDField(read_only=True)


class RepositorySerializer(serializers.ModelSerializer):
    """
    Serialize a repository
    """
    enabled = serializers.BooleanField(read_only=True)
    git_clone_url = serializers.SerializerMethodField()
    workers = WorkerSerializer(many=True, read_only=True)
    authorized_users = serializers.SerializerMethodField(read_only=True)

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_git_clone_url(self, repository):
        # This check avoid to set git_clone_url when this serializer is used to list multiple
        # repositories because self.instance value would be a list, even if the request user is internal
        if self.context['request'].user.is_internal and isinstance(self.instance, Repository) and self.instance.enabled:
            url = urllib.parse.urlparse(self.instance.url)
            return f"https://oauth2:{repository.credentials.token}@{url.netloc}{url.path}"
        return None

    def get_authorized_users(self, repo) -> int:
        count = getattr(repo, 'authorized_users', None)
        if count is None:
            count = repo.memberships.count()
        return count

    class Meta:
        model = Repository
        fields = (
            'id',
            'url',
            'enabled',
            'git_clone_url',
            'workers',
            'authorized_users',
        )
        extra_kwargs = {
            'id': {'read_only': True},
            'url': {'read_only': True}
        }


class WorkerActivitySerializer(serializers.ModelSerializer):
    """
    Serialize a worker activity on an element
    """
    state = EnumField(WorkerActivityState)
    element_id = serializers.UUIDField()
    process_id = serializers.PrimaryKeyRelatedField(
        queryset=Process.objects.all(),
        # Avoid loading up to 1000 processes when opening this endpoint in a browser
        style={'base_template': 'input.html'},
    )
    configuration_id = serializers.UUIDField(read_only=True, allow_null=True, required=False)
    worker_version_id = serializers.UUIDField(read_only=True)

    class Meta:
        model = WorkerActivity
        fields = (
            'element_id',
            'process_id',
            'configuration_id',
            'worker_version_id',
            'created',
            'updated',
            'state',
        )
        read_only_fields = (
            'created',
            'updated',
            'configuration_id',
            'worker_version_id',
        )


class WorkerConfigurationListSerializer(serializers.ModelSerializer):
    configuration = serializers.DictField(allow_empty=False)

    class Meta:
        model = WorkerConfiguration
        fields = (
            'id',
            'name',
            'configuration',
            'archived'
        )

    def validate(self, data):
        # Check unique constraints
        name = data.get('name')
        configuration = data.get('configuration')
        configuration_hash = hash_object(configuration)
        data['archived'] = data.get('archived', False)
        worker = self.context.get('worker')
        # Since WorkerConfigurationSerializer is built from this serializer, this validation function can be triggered by a call
        # to the WorkerConfigurationRetrieve API endpoint, updating an existing configuration. In this case, there is no worker
        # cached property and this validation is not necessary anyway because the name and configuration fields are read_only.
        if not worker:
            return data
        data['worker_id'] = worker.id
        existing_configurations = list(worker.configurations.filter(Q(configuration_hash=configuration_hash) | Q(name=name)))
        errors = defaultdict(list)
        for config in existing_configurations:
            if configuration_hash == config.configuration_hash:
                errors['configuration'].append("A worker configuration with this configuration already exists for this worker.")
                errors['id'] = config.id
            if name == config.name:
                errors['name'].append("A worker configuration with this name already exists for this worker.")
        # Due to the defaultdict, writing e.g. len(errors['configuration']) > 1 would add {'configuration': []} to
        # the errors dict if the key is not already in the dict.
        if len(errors.get('name', [])) > 1 or len(errors.get('configuration', [])) > 1:
            # If there is more than one configuration with the same name and/or configuration as the one the user
            # is trying to create, raise a MultipleObjectsReturned exception because this shouldn't be possible
            # due to the unicity constraints. There can however be more than one worker configurations in
            # existing_configurations, if both a worker configuration with the same name and the same configuration
            # exist; but checking the length of existing_configurations wouldn't differentiate between this case
            # and an actual duplicated worker configuration.
            raise WorkerConfiguration.MultipleObjectsReturned
        if errors:
            raise ValidationError(errors)
        return data


class WorkerConfigurationSerializer(WorkerConfigurationListSerializer):
    configuration = serializers.DictField(allow_empty=False, read_only=True)

    class Meta(WorkerConfigurationListSerializer.Meta):
        # Only allow updating `archived`
        read_only_fields = ('id', 'name', 'configuration')


class WorkerConfigurationExistsErrorSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="UUID of an existing worker configuration, if the error comes from a duplicate configuration.")
    configuration = serializers.ListField(child=serializers.CharField(), required=False, help_text="Configuration error message.")
    name = serializers.ListField(child=serializers.CharField(), required=False, help_text="Name error message.")


class WorkerStatisticsSerializer(serializers.Serializer):
    """
    Serialize activity statistics of a worker version
    """
    worker_version_id = serializers.UUIDField(read_only=True)
    configuration_id = serializers.UUIDField(read_only=True, allow_null=True)
    queued = serializers.IntegerField(read_only=True)
    started = serializers.IntegerField(read_only=True)
    processed = serializers.IntegerField(read_only=True)
    error = serializers.IntegerField(read_only=True)
