from collections.abc import Iterable, Iterator, Sized

from django.db.models import CharField, Func
from django.urls import reverse
from rest_framework import serializers

from arkindex.documents.models import Element, ElementPath


def build_absolute_url(element, request, name, id_argument='pk', **kwargs):
    """
    Build an absolute URL for a specified view using the element ID. Used by IIIF serializers.
    """
    kwargs[id_argument] = str(element.id)
    return request.build_absolute_uri(reverse(name, kwargs=kwargs))


def build_tree(tree, *, corpus, type):
    """
    Build Element and ElementPath instances from a tree described by
    a dict associating elements with their direct ancestors.
    Should only be used for unit tests.

    {'child': ['mother', 'father'], 'grandchild': 'child'}

    Returns a dict associating element names with created Elements.
    """
    assert isinstance(tree, dict)

    def parse_value(val):
        if isinstance(val, str):
            return [val]
        return val

    # List all elements
    # Pick all the keys, then add elements in the tree's values
    # Removes the need to add 'element': [] for root elements
    names = set(tree.keys()).union(*[parse_value(val) for val in tree.values()])
    # Create elements
    elements = {
        name: Element.objects.create(
            corpus=corpus,
            name=name,
            type=type,
        )
        for name in names
    }
    # Create empty paths for every element; add_parent will handle removing them if needed.
    ElementPath.objects.bulk_create([
        ElementPath(element=element, path=[], ordering=0)
        for element in elements.values()
    ])
    # Create paths
    for child_name, value in tree.items():
        parent_names = parse_value(value)
        for parent_name in parent_names:
            elements[child_name].add_parent(elements[parent_name])

    return elements


def bounding_box(linestring):
    """
    Returns a 4-tuple (x, y, width, height) for the bounding box of a LineString or LinearRing
    """
    all_x, all_y = zip(*linestring.coords)
    x, y = min(all_x), min(all_y)
    width, height = max(all_x) - x, max(all_y) - y
    return int(x), int(y), int(width), int(height)


def polygon_outside_image(img, linestring):
    x, y, w, h = bounding_box(linestring)
    return (x > img.width or x + w > img.width or y > img.height or y + h > img.height)


class BulkMap(Sized, Iterable):
    """
    Apply a function on the full list of items as soon as an iterator is requested.

    Useful to perform complex operations on a QuerySet with pagination
    without breaking through DRF's and Django's layers.
    """

    def __init__(self, func, iterable):
        assert callable(func), 'func must be callable'
        assert isinstance(iterable, Iterable), 'iterable must implement __iter__'
        assert isinstance(iterable, Sized), 'iterable must implement __len__'
        self.func = func
        self.iterable = iterable
        self._cached_length = None

    def __getitem__(self, key):
        assert isinstance(key, slice), 'BulkMap only supports slicing'
        return BulkMap(self.func, self.iterable[key])

    def __len__(self):
        if self._cached_length is None:
            # Proxy __len__ without calling the function
            if hasattr(self.iterable, 'count'):
                self._cached_length = self.iterable.count()
            else:
                self._cached_length = len(self.iterable)
        return self._cached_length

    def __iter__(self):
        return iter(self.func(list(self.iterable)))


class CounterIterator(Sized, Iterator):
    """
    Iterate over any iterator while keeping count of how many items were iterated on.
    This allows accessing the total amount of items in an iterator once a full iteration is completed
    without storing any item in memory.
    """

    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.count = 0

    def __next__(self):
        # This can fail with StopIteration if the iteration ends.
        item = next(self.iterator)
        # It didn't fail, so we can increment the counter.
        self.count += 1
        return item

    def __len__(self):
        return self.count


class RTrimChr(Func):
    """
    Django's RTrim function only allows trimming spaces.
    This function allows trimming any character, not only spaces.
    """
    function = 'RTRIM'
    arity = 2
    output_field = CharField()


class WorkerRunOrVersionValidator(object):
    """
    A generic DRF validator to facilitate the transition between worker_version_id and worker_run_id.
    In the API, only one of the two fields is accepted.
    """
    requires_context = True

    def __init__(self, worker_version_field='worker_version_id', worker_run_field='worker_run_id'):
        self.worker_version_field = worker_version_field
        self.worker_run_field = worker_run_field

    def __call__(self, attrs, serializer):
        # Validators like this one run pretty late in DRF's validation process and the attributes we get
        # are not the ones defined in the serializer: they are influenced by the `source` of the fields.
        # The proper way to define a worker run ID is to use a PrimaryKeyRelatedField with source='worker_run',
        # which retrieves the worker run itself and will store it in the source.
        # The client sends an UUID as `worker_run_id`, but the PrimaryKeyRelatedField finds it and stores it
        # in `worker_run`.  Storing a WorkerRun instance inside `worker_run_id` causes a Django ValidationError
        # because a WorkerRun is not a valid UUID.
        worker_version_key = serializer.fields[self.worker_version_field].source
        worker_run_key = serializer.fields[self.worker_run_field].source
        if attrs.get(worker_version_key) is not None and attrs.get(worker_run_key) is not None:
            raise serializers.ValidationError(f'Only one of `{self.worker_version_field}` and `{self.worker_run_field}` may be set.')


class ConditionalUniqueValidator(serializers.UniqueTogetherValidator):
    """
    A UniqueTogetherValidator that only runs when a condition is fulfilled.
    This is useful to validate partial unique constraints, which use `condition=Q(…)`.

    The `condition` keyword argument should be a function that takes a dict to validate
    and returns True if the unique constraint should be checked and False otherwise.
    The dict is the same as found in `ModelSerializer.validate(value)` as the `value`.
    """

    def __init__(self, *args, condition=None, **kwargs):
        assert callable(condition), 'The condition should be a function.'
        self.condition = condition
        super().__init__(*args, **kwargs)

    def __call__(self, attrs, serializer):
        if not self.condition(attrs):
            return
        return super().__call__(attrs, serializer)
