from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import serializers, status
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.generics import ListCreateAPIView, RetrieveAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.response import Response

from arkindex.project.mixins import TrainingModelMixin
from arkindex.project.permissions import IsVerified
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.training.serializers import (
    CreateModelErrorResponseSerializer,
    ModelSerializer,
    ModelVersionCreateSerializer,
    ModelVersionSerializer,
)
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListModelVersions',
        description=(
            'List available versions of a Machine Learning model.\n\n'
            'Requires a **guest** access to the model.'
        ),
    ),
    post=extend_schema(
        operation_id='CreateModelVersion',
        description=(
            'Create a new version for a Machine Learning model.\n\n'
            'Requires a **contributor** access to the model.'
        ),
        responses={
            200: ModelVersionCreateSerializer, 400: None, 403: None
        },
    ),
)
class ModelVersionsList(TrainingModelMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ModelVersionCreateSerializer
    queryset = Model.objects.none()

    def get_queryset(self):
        return Model.objects.filter(pk=self.kwargs['pk'])

    def perform_create(self, serializer):
        if not self.model_access_rights or self.model_access_rights < Role.Contributor.value:
            raise PermissionDenied(detail='You need a Contributor access to the model to create a new version.')
        serializer.save()

    @cached_property
    def model(self):
        return get_object_or_404(self.get_queryset(), id=self.kwargs['pk'])

    @cached_property
    def model_access_rights(self):
        return get_max_level(self.request.user, self.model)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['model'] = self.model
        context['model_rights'] = self.model_access_rights
        return context

    def filter_queryset(self, queryset):
        filters = Q(model=self.kwargs['pk'])

        access_level = self.model_access_rights
        if not access_level or access_level < Role.Guest.value:
            raise PermissionDenied(detail='You need a guest access to the model to list its versions.')

        # Guest access allow only versions with a set tag and state = available
        if access_level < Role.Contributor.value:
            filters &= Q(tag__isnull=False, state=ModelVersionState.Available)

        # Order version by creation date
        queryset = ModelVersion.objects.filter(filters).order_by('-created')

        return super().filter_queryset(queryset)


@extend_schema(tags=['training'])
@extend_schema_view(
    put=extend_schema(
        description=(
            'Update a version of a Machine Learning model.\n\n'
            'Requires a **contributor** access to the model.'
        ),
    ),
    patch=extend_schema(
        description='Partially update a version of a Machine Learning model.\n\nRequires a **contributor** access to the model.'
    ),
    delete=extend_schema(
        description='Delete a model version.\n\nRequires an **admin** access on the related model.'
    )
)
class ModelVersionsRetrieve(TrainingModelMixin, RetrieveUpdateDestroyAPIView):
    """Retrieve a version of Machine Learning model
    """
    permission_classes = (IsVerified, )
    serializer_class = ModelVersionSerializer
    queryset = ModelVersion.objects.none()

    def get_queryset(self):
        return ModelVersion.objects \
            .filter(id=self.kwargs['pk']) \
            .select_related('model')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['model'] = self.model
        return context

    @cached_property
    def model(self):
        return get_object_or_404(self.get_queryset(), id=self.kwargs['pk']).model

    @cached_property
    def model_access_rights(self):
        return get_max_level(self.request.user, self.model)

    def check_object_permissions(self, request, model_version):
        access_level = self.model_access_rights

        if model_version.state == ModelVersionState.Available and model_version.tag is not None:
            needed_level, error_msg = Role.Guest.value, 'Guest'
        else:
            needed_level, error_msg = Role.Contributor.value, 'Contributor'

        if not access_level or access_level < needed_level:
            raise PermissionDenied(detail=f'You need a {error_msg} access to the model to retrieve this version.')
        return super().check_object_permissions(request, model_version)

    def perform_update(self, serializer):
        access_level = self.model_access_rights
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail='You need a Contributor access to the model to edit this version.')
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        access_level = self.model_access_rights
        if not access_level or access_level < Role.Admin.value:
            raise PermissionDenied(detail='You need an Admin access to the model to destroy this version.')

        super().perform_destroy(instance)


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListModels',
        description=(
            'List available Machine Learning models.'
        ),
        responses={
            200: ModelSerializer,
        },
        parameters=[
            OpenApiParameter(
                'name',
                description='Filter models whose name contains the given string (case insensitive)',
                required=False,
            ),
        ]
    ),
    post=extend_schema(
        description=(
            'Create a new Machine Learning model.'
        ),
        responses={
            200: ModelSerializer,
            400: CreateModelErrorResponseSerializer,
            403: None
        },
    ),
)
class ModelsList(TrainingModelMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ModelSerializer

    def get_queryset(self):
        filters = Q()
        if 'name' in self.request.query_params:
            filters &= Q(name__icontains=self.request.query_params['name'])

        return self.readable_models.filter(filters).order_by('name')

    def perform_create(self, serializer):
        model_name = serializer.validated_data.get('name')
        existing_model = Model.objects.using('default').filter(name=model_name).first()
        if existing_model:
            if self.has_read_access(existing_model):
                raise ValidationError({
                    'id': str(existing_model.id),
                    'name': 'A model with this name already exists',
                })
            else:
                raise PermissionDenied()
        return serializer.save()


@extend_schema(tags=['training'])
@extend_schema_view(
    get=extend_schema(
        operation_id='DownloadModelVersion',
        description=(
            'Retrieve a download url for a model_version. A secret token is required to have access to the model version.'
            'This endpoint does **not** require to be logged in.'
        ),
        responses={
            302: serializers.Serializer,
        },
        parameters=[
            OpenApiParameter(
                'token',
                description='SHA256 hash used for authentication and check access to the model version.',
                required=True,
            ),
        ]
    ),
)
class ModelVersionDownload(TrainingModelMixin, RetrieveAPIView):
    queryset = ModelVersion.objects.all()

    def check_object_permissions(self, request, model_version):
        token = model_version.build_authentication_token_hash()
        if token != request.query_params['token']:
            raise PermissionDenied()

    def get(self, request, *args, **kwargs):
        return Response(status=status.HTTP_302_FOUND, headers={'Location': self.get_object().s3_url})
