
import re

from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from rest_framework.validators import UniqueTogetherValidator

from arkindex.project.mixins import TrainingModelMixin
from arkindex.project.serializer_fields import EnumField
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


def _model_from_context(serializer_field):
    return serializer_field.context.get('model')


_model_from_context.requires_context = True


class ModelLightSerializer(serializers.ModelSerializer):
    name = serializers.CharField(max_length=100)

    class Meta:
        model = Model
        fields = ('id', 'name')


class ModelSerializer(TrainingModelMixin, ModelLightSerializer):
    # Actually define the field to avoid the field-level automatically generated UniqueValidator
    rights = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = Model
        fields = ModelLightSerializer.Meta.fields + ('created', 'updated', 'description', 'rights')

    def create(self, validated_data):
        instance = super().create(validated_data)

        # Create admin membership for creator user
        instance.memberships.create(
            user=self.context['request'].user,
            level=Role.Admin.value
        )
        return instance

    @extend_schema_field(serializers.ListField(child=serializers.ChoiceField(['read', 'write', 'admin'])))
    def get_rights(self, model):
        # to avoid a stale read on get_max_level when creating a model (because the corresponding membership is
        # created at the same time as the model) if a model is being created we return these rights without
        # actually checking them.
        if self.context['request'].method == 'POST':
            return ['read', 'write', 'admin']

        level = get_max_level(self.context['request'].user, model)
        rights = ['read']
        if level >= Role.Contributor.value:
            rights.append('write')
        if level >= Role.Admin.value:
            rights.append('admin')
        return rights


class CreateModelErrorResponseSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="UUID of an existing model, if the error comes from a duplicate name.")
    name = serializers.CharField(required=False, help_text="Name of an existing model, if the error comes from a duplicate name.")
    detail = serializers.CharField(required=False, help_text="A generic error message when an error occurs outside of a specific field.")


class ModelVersionLightSerializer(serializers.ModelSerializer):
    """
    Serialize a simple model version
    """

    model = ModelLightSerializer(default=_model_from_context)
    state = EnumField(ModelVersionState)
    configuration = serializers.JSONField(style={'base_template': 'textarea.html'})

    class Meta:
        model = ModelVersion
        fields = ('id', 'model', 'tag', 'state', 'size', 'configuration')
        read_only_fields = ('id', 'model', 'tag', 'state', 'size', 'configuration')


class ModelVersionSerializer(TrainingModelMixin, ModelVersionLightSerializer):
    description = serializers.CharField(allow_blank=True, style={'base_template': 'textarea.html'})
    s3_url = serializers.SerializerMethodField()
    model = serializers.HiddenField(default=_model_from_context)

    class Meta:
        model = ModelVersion
        fields = ModelVersionLightSerializer.Meta.fields + ('model_id', 'parent', 'description', 'hash', 'archive_hash', 's3_url')
        read_only_fields = ('id', 'model', 'parent', 'state', 'hash', 'archive_hash', 'size', 's3_url')
        validators = [
            UniqueTogetherValidator(
                queryset=ModelVersion.objects.filter(tag__isnull=False),
                fields=['model', 'tag'],
                message='A version for this model with this tag already exists.'
            )
        ]

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_url(self, obj):
        return obj.s3_url

    def validate_state(self, state):
        # If User wants to update the state to Available, check the file on s3
        if state == ModelVersionState.Available:
            try:
                # Save after hash has been checked, no need to do it twice
                self.instance.perform_check(save=False, raise_exc=True)
                self.instance.check_hash(raise_exc=True)
            except (AssertionError, ValueError) as e:
                raise ValidationError(str(e))
        return state

    def validate_hash(self, hash):
        existing_modelversion = self.context['model'].versions.filter(hash=hash).first()
        if existing_modelversion:
            if existing_modelversion.state != ModelVersionState.Available:
                raise ValidationError(ModelVersionCreateSerializer(existing_modelversion).data)
            else:
                raise ValidationError(detail="A version for this model with this hash already exists.")
        return hash


class ModelVersionCreateSerializer(ModelVersionSerializer):
    """
    Create a new Model Version by providing the hash and the size of the archive
    """
    parent = serializers.PrimaryKeyRelatedField(queryset=ModelVersion.objects.none(), default=None)
    hash = serializers.RegexField(re.compile(r'[0-9A-Fa-f]{32}'), min_length=32, max_length=32)
    archive_hash = serializers.RegexField(re.compile(r'[0-9A-Fa-f]{32}'), min_length=32, max_length=32)
    description = serializers.CharField(required=False, style={'base_template': 'textarea.html'})
    size = serializers.IntegerField(min_value=0)
    s3_put_url = serializers.SerializerMethodField()
    state = EnumField(ModelVersionState, default=ModelVersionState.Created, read_only=True)
    configuration = serializers.JSONField(required=False, decoder=None, encoder=None, style={'base_template': 'textarea.html'})
    tag = serializers.CharField(allow_null=True, max_length=50, required=False, default=None)

    class Meta(ModelVersionSerializer.Meta):
        fields = ModelVersionSerializer.Meta.fields + ('s3_put_url', 'created', )
        read_only_fields = ModelVersionSerializer.Meta.read_only_fields + ('s3_put_url', 'created', )

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_put_url(self, obj):
        return obj.s3_put_url

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # The API adds the provided model to the context
        if self.context.get('model') is not None:
            self.fields['parent'].queryset = ModelVersion.objects.filter(model_id=self.context['model'].id)
            # If user doesn't have a contributor access to the model, don't show s3_url and s3_put_url
            access_level = self.context.get('model_rights')
            if not access_level or access_level < Role.Contributor.value:
                del self.fields['s3_url']
                del self.fields['s3_put_url']
