

from unittest.mock import patch
from uuid import uuid4

from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Group, Right, Role, User


def _format_datetime(date):
    return str(date.isoformat().replace('+00:00', 'Z')).replace(' ', 'T')


def _deserialize_model(model, access_rights):
    return {
        'id': str(model.id),
        'created': _format_datetime(model.created),
        'updated': _format_datetime(model.updated),
        'name': model.name,
        'description': model.description,
        'rights': access_rights
    }


def _deserialize_model_version(model_version):
    # Only stringify if we get an ID
    parent = str(model_version.parent) if model_version.parent else model_version.parent
    return {
        'id': str(model_version.id),
        'model_id': str(model_version.model_id),
        'parent': parent,
        'description': model_version.description,
        'tag': model_version.tag,
        'hash': model_version.hash,
        'archive_hash': model_version.archive_hash,
        'state': model_version.state.value,
        'size': model_version.size,
        'configuration': model_version.configuration,
        's3_url': model_version.s3_url,
    }


class TestModelAPI(FixtureAPITestCase):
    """
    Test model and model version api
    """

    @classmethod
    def setUpTestData(cls):
        r"""We use a simple rights configuration for those tests

            User1     User2       User3
              |        /           /
              |      100          /
              |       |          /
             100   Group1      10
              |     / \        /
              |   50   100    /
              |  /       \   /
            Model1      Model2
        """
        super().setUpTestData()

        # Create there models
        cls.model1 = Model.objects.create(name="First Model")
        cls.model2 = Model.objects.create(name="Second Model")
        cls.model3 = Model.objects.create(name="Third Model")

        # Create some Model Versions
        cls.model_version1 = ModelVersion.objects.create(model=cls.model1, hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", archive_hash="5eca9bd3eb07c006cd43ae48dfde7fd3", size=8)
        cls.model_version2 = ModelVersion.objects.create(model=cls.model1, description="some description", tag="tagged", configuration={"n_epochs": '10'}, hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb", archive_hash="8b4f9ea16de4bcf5bbfc0ff1ea237934", size=8)
        cls.model_version3 = ModelVersion.objects.create(model=cls.model2, state="available", tag="tagged", hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbba", archive_hash="a501b1ae6f6bc833551245f8328590b8", size=8)
        cls.model_version4 = ModelVersion.objects.create(model=cls.model2, description="some description", tag="taggedv2", configuration={"n_epochs": '10'}, hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbaa", archive_hash="d67459a391b228cced507f068d4a570a", size=8)
        cls.model_version5 = ModelVersion.objects.create(model=cls.model1, description="some description", tag="available_version", state='available', configuration={"n_epochs": 10}, hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbdd", archive_hash="3d87fd18e03fcbbfb64a381ad9472596", size=8)

        # Create three users with different access rights on each model
        cls.user1 = User.objects.create(email='user1@test.test', display_name='User 1', verified_email=True)
        cls.user2 = User.objects.create(email='user2@test.test', display_name='User 2', verified_email=True)
        cls.user3 = User.objects.create(email='user3@test.test', display_name='User 3', verified_email=True)

        # Create the group
        cls.group1 = Group.objects.create(name="Group1")

        # Access rights
        Right.objects.bulk_create([
            Right(user=cls.user1, content_object=cls.model1, level=100),
            Right(user=cls.user2, content_object=cls.group1, level=100),
            Right(user=cls.user3, content_object=cls.model2, level=10),
            Right(group=cls.group1, content_object=cls.model1, level=50),
            Right(group=cls.group1, content_object=cls.model2, level=100),
        ])

    def build_model_version_create_request(self):
        return {
            'hash': '94274e84f3de91d1645b1e082b5f3466',
            'archive_hash': '0958a74b060a89fc38318a9a96aef32a',
            'size': 8
        }

    def build_model_version_update_request(self):
        return {
            'state': ModelVersionState.Available.value,
            'description': 'description',
            'configuration': {"hello": "this is me"},
        }

    def test_create_model_version_requires_verified(self):
        user = User.objects.create(email='not_verified@mail.com', display_name='Not Verified', verified_email=False)
        self.client.force_login(user)
        request = self.build_model_version_create_request()
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:model-versions", kwargs={"pk": str(self.model2.id)}), request, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_model_version_requires_contributor(self):
        """
        Can't create model version as guest
        """
        self.client.force_login(self.user1)
        request = self.build_model_version_create_request()
        with self.assertNumQueries(7):
            response = self.client.post(reverse("api:model-versions", kwargs={"pk": str(self.model2.id)}), request, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the model to create a new version."})

    @patch('arkindex.training.models.logger')
    @patch('arkindex.project.aws.S3FileMixin.s3_object')
    def test_model_version_check_hash_ignores_multipart(self, s3_object_mock, logger_mock):
        s3_object_mock.e_tag = '"badbadbadbad-5"'
        version = ModelVersion.objects.create(model_id=self.model1.id, archive_hash='huehuehuehue', size=8, hash='huehuehue')

        version.check_hash(save=True, raise_exc=True)
        self.assertEqual(version.state, ModelVersionState.Available)
        self.assertEqual(logger_mock.warning.call_count, 1)

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_model_version_creation_no_tag(self, s3_presigned_url_mock):
        """
        Creates a new model version without setting a tag
        """
        s3_presigned_url_mock.return_value = 'http://s3/upload_put_url'
        self.client.force_login(self.user1)
        request = self.build_model_version_create_request()

        fake_now = timezone.now()
        # To mock the creation date
        with patch('django.utils.timezone.now') as mock_now:
            mock_now.return_value = fake_now
            with self.assertNumQueries(8):
                response = self.client.post(reverse('api:model-versions', kwargs={"pk": str(self.model1.id)}), request, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        df = ModelVersion.objects.get(id=data['id'])

        self.assertDictEqual(
            data,
            {
                'id': str(df.id),
                'model_id': str(self.model1.id),
                'parent': None,
                'description': '',
                'state': ModelVersionState.Created.value,
                'configuration': {},
                'tag': None,
                'size': request['size'],
                'hash': request['hash'],
                'archive_hash': request['archive_hash'],
                'created': _format_datetime(fake_now),
                's3_url': s3_presigned_url_mock.return_value,
                's3_put_url': s3_presigned_url_mock.return_value
            }
        )

    def test_create_model_version_blank_tag_model(self):
        """
        Raise 400 when creating a model version with a blank tag
        """
        self.client.force_login(self.user1)
        request = self.build_model_version_create_request()
        request['tag'] = ''
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:model-versions', kwargs={"pk": str(self.model1.id)}), request, format='json')

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"tag": ["This field may not be blank."]})

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_model_version_creation_with_tag(self, s3_presigned_url_mock):
        """
        Creates a new model version with a tag
        """
        self.client.force_login(self.user1)
        s3_presigned_url_mock.return_value = 'http://s3/upload_put_url'
        request = {
            'tag': 'TAG',
            'hash': '05a5cdbaf05d3b6cc51fcb173d0057c0',
            'archive_hash': '05a5cdbaf05d3b6cc51fcb173d0057c0',
            'size': 8,
        }
        fake_now = timezone.now()
        # To mock the creation date
        with patch('django.utils.timezone.now') as mock_now:
            mock_now.return_value = fake_now
            with self.assertNumQueries(9):
                response = self.client.post(reverse('api:model-versions', kwargs={"pk": str(self.model1.id)}), request, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        df = ModelVersion.objects.get(id=data['id'])

        self.assertDictEqual(
            data,
            {
                'id': str(df.id),
                'model_id': str(self.model1.id),
                'parent': None,
                'description': '',
                'state': ModelVersionState.Created.value,
                'configuration': {},
                'tag': request['tag'],
                'size': request['size'],
                'hash': request['hash'],
                'archive_hash': request['archive_hash'],
                'created': _format_datetime(fake_now),
                's3_url': s3_presigned_url_mock.return_value,
                's3_put_url': s3_presigned_url_mock.return_value
            }
        )

    def test_create_model_version_unique_tag_model(self):
        """
        Raises 400 when creating a model version that already exists, same model_id and tag
        """
        self.client.force_login(self.user1)
        request = self.build_model_version_create_request()
        request['tag'] = self.model_version2.tag
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:model-versions', kwargs={"pk": str(self.model1.id)}), request, format='json')

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A version for this model with this tag already exists."]})

    def test_create_model_version_unique_hash_model_available(self):
        """
        Raises 400 when creating a model version that already exists, same model_id and hash and state==Available
        """
        self.client.force_login(self.user1)
        request = {'tag': 'production', 'hash': self.model_version5.hash, 'archive_hash': self.model_version5.archive_hash, 'size': 32}
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:model-versions', kwargs={"pk": str(self.model1.id)}), request, format='json')

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"hash": ["A version for this model with this hash already exists."]})

    def test_create_model_version_unique_hash_model_not_available(self):
        """
        Raises 400 when creating a model version that already exists, same model_id and hash but state != Available. Returns usual payload with s3_put_url of the match
        """
        self.client.force_login(self.user1)
        request = {'tag': 'production', 'hash': self.model_version2.hash, 'size': 32}
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:model-versions', kwargs={"pk": str(self.model1.id)}), request, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json()['hash'],
            {
                'id': str(self.model_version2.id),
                'model_id': str(self.model1.id),
                'parent': str(self.model_version2.parent),
                'description': self.model_version2.description,
                'state': self.model_version2.state.value,
                'configuration': self.model_version2.configuration,
                'tag': self.model_version2.tag,
                'created': _format_datetime(self.model_version2.created),
                'size': str(self.model_version2.size),
                'hash': self.model_version2.hash,
                'archive_hash': self.model_version2.archive_hash,
                's3_url': self.model_version2.s3_url,
                's3_put_url': self.model_version2.s3_put_url
            }
        )

    @patch('arkindex.project.aws.s3.Object')
    @patch('arkindex.project.aws.S3FileMixin.exists')
    def test_partial_update_model_version_requires_contributor(self, exists, s3_object):
        """
        Can't partial update a model version as guest
        """
        s3_object().content_length = self.model_version3.size
        s3_object().e_tag = self.model_version3.archive_hash
        exists.return_value = True
        self.client.force_login(self.user3)
        with self.assertNumQueries(9):
            response = self.client.patch(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}), {'state': 'available'})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the model to edit this version."})

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    @patch('arkindex.project.aws.S3FileMixin.exists')
    @patch('arkindex.project.aws.s3.Object')
    def test_partial_update_model_version(self, s3_object, exists, s3_presigned_url):
        """PartialUpdate the state to available and the model version's description, tag and configuration
        """
        s3_presigned_url.return_value = "http://s3/upload_url"
        exists.return_value = True
        s3_object().content_length = self.model_version1.size
        s3_object().e_tag = self.model_version1.archive_hash
        self.client.force_login(self.user1)
        request = {
            "description" : "A very long description",
            "tag" : "prod",
            "state" : "available",
            "configuration": {"parameter1": "value1"}
        }
        with self.assertNumQueries(10):
            response = self.client.patch(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}), request, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.model_version1.id),
            "model_id": str(self.model_version1.model.id),
            "parent": self.model_version1.parent,
            "description": request.get('description'),
            "tag": request.get('tag'),
            "hash": self.model_version1.hash,
            "archive_hash": self.model_version1.archive_hash,
            "state": "available",
            "size": self.model_version1.size,
            "configuration": request.get('configuration'),
            "s3_url": "http://s3/upload_url",
        })

    @patch('arkindex.project.aws.S3FileMixin.exists')
    def test_partial_update_model_version_archive_not_present(self, exists):
        """Raises 400 when archive is not found while updating partially the state
        """
        self.client.force_login(self.user1)
        exists.return_value = False
        with self.assertNumQueries(7):
            response = self.client.patch(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}), {
                'state': 'available'
            })

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": ['Archive has not been uploaded']})

    @patch('arkindex.project.aws.S3FileMixin.exists')
    @patch('arkindex.project.aws.s3.Object')
    def test_partial_update_model_version_wrong_size(self, s3_object, exists):
        """Raises 400 when archive sizes don't match while updating partially the state
        """
        exists.return_value = True
        s3_object().content_length = self.model_version1.size + 1
        self.client.force_login(self.user1)
        with self.assertNumQueries(7):
            response = self.client.patch(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}), {
                'state': 'available',
            })

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": [f'Uploaded file size is {self.model_version1.size + 1} bytes, expected {self.model_version1.size} bytes']})

    @patch('arkindex.project.aws.S3FileMixin.exists')
    @patch('arkindex.project.aws.s3.Object')
    def test_partial_update_model_version_wrong_hash(self, s3_object, exists):
        """Raises 400 when archive hashes don't match while updating partially the state
        """
        exists.return_value = True
        s3_object().content_length = self.model_version1.size
        s3_object().e_tag = f'"{self.model_version2.hash}"'
        self.client.force_login(self.user1)
        with self.assertNumQueries(8):
            response = self.client.patch(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}), {
                'state': 'available',
            })

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": ['MD5 hashes do not match']})

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    @patch('arkindex.project.aws.S3FileMixin.exists')
    @patch('arkindex.project.aws.s3.Object')
    def test_partial_update_model_version_unique(self, s3_object, exists, s3_presigned_url):
        """
        Raises 400 when partial updating a model version and ending up with a version that has the same model_id and tag as another one.
        """
        s3_presigned_url.return_value = "http://s3/upload_put_url"
        exists.return_value = True
        s3_object().content_length = self.model_version1.size
        s3_object().e_tag = self.model_version1.archive_hash
        self.client.force_login(self.user1)
        request = {
            "tag": self.model_version2.tag,
            "state": "available",
        }
        with self.assertNumQueries(9):
            response = self.client.patch(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}), request, format='json')

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A version for this model with this tag already exists."]})

    @patch('arkindex.project.aws.s3.Object')
    @patch('arkindex.project.aws.S3FileMixin.exists')
    def test_update_model_version_requires_contributor(self, exists, s3_object):
        """
        Can't update a model version with guest access rights to the model
        """
        s3_object().content_length = self.model_version3.size
        s3_object().e_tag = self.model_version3.archive_hash
        exists.return_value = True
        self.client.force_login(self.user3)
        request = self.build_model_version_update_request()
        with self.assertNumQueries(9):
            response = self.client.put(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version3.id)}), request, format='json')
        # self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the model to edit this version."})

    def test_update_model_version_requires_all_parameters(self):
        """
        Update endpoint requires every parameter
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(7):
            response = self.client.put(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}), {})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'description': ['This field is required.'], 'state': ['This field is required.'], 'configuration': ['This field is required.']})

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    @patch('arkindex.project.aws.S3FileMixin.exists')
    @patch('arkindex.project.aws.s3.Object')
    def test_update_model_version_missing_parameters(self, s3_object, exists, s3_presigned_url):
        """
        Update endpoint requires every parameter, not only state
        """
        s3_presigned_url.return_value = "http://s3/upload_url"
        exists.return_value = True
        s3_object().content_length = self.model_version1.size
        s3_object().e_tag = self.model_version1.archive_hash

        self.client.force_login(self.user1)
        params = self.build_model_version_update_request()
        for param, value in params.items():
            response = self.client.put(reverse("api:model-version-retrieve", kwargs={"pk": str(self.model_version1.id)}), {param: value}, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), {other_param: ['This field is required.'] for other_param in params.keys() if other_param != param}, f'{param}')

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    @patch('arkindex.project.aws.S3FileMixin.exists')
    @patch('arkindex.project.aws.s3.Object')
    def test_update_model_version(self, s3_object, exists, s3_presigned_url):
        """Update the state to available and updates the model version's description, tag and configuration
        """
        s3_presigned_url.return_value = "http://s3/upload_url"
        exists.return_value = True
        s3_object().content_length = self.model_version2.size
        s3_object().e_tag = self.model_version2.archive_hash
        self.client.force_login(self.user1)

        request = {
            "description" : "A very long description",
            "state": "available",
            "configuration" : {"parameter1": "value1"}
        }
        with self.assertNumQueries(10):
            response = self.client.put(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version2.id)}), request, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.model_version2.id),
            "model_id": str(self.model_version2.model.id),
            "parent": self.model_version2.parent,
            "description": request.get('description'),
            "tag": self.model_version2.tag,
            "hash": self.model_version2.hash,
            "archive_hash": self.model_version2.archive_hash,
            "state": "available",
            "size": self.model_version2.size,
            "configuration": request.get('configuration'),
            "s3_url": "http://s3/upload_url",
        })

    @patch('arkindex.project.aws.S3FileMixin.exists')
    def test_update_model_version_archive_not_present(self, exists):
        """Raises 400 when archive is not found
        """
        self.client.force_login(self.user1)
        exists.return_value = False
        request = self.build_model_version_update_request()
        with self.assertNumQueries(7):
            response = self.client.put(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}), request, format='json')

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": ['Archive has not been uploaded']})

    @patch('arkindex.project.aws.S3FileMixin.exists')
    @patch('arkindex.project.aws.s3.Object')
    def test_update_model_version_wrong_size(self, s3_object, exists):
        """Raises 400 when archive sizes don't match
        """
        exists.return_value = True
        s3_object().content_length = self.model_version1.size + 1
        self.client.force_login(self.user1)
        request = self.build_model_version_update_request()
        with self.assertNumQueries(7):
            response = self.client.put(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}), request, format='json')

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": [f'Uploaded file size is {self.model_version1.size + 1} bytes, expected {self.model_version1.size} bytes']})

    @patch('arkindex.project.aws.S3FileMixin.exists')
    @patch('arkindex.project.aws.s3.Object')
    def test_update_model_version_wrong_hash(self, s3_object, exists):
        """Raises 400 when archive hashes don't match
        """
        exists.return_value = True
        s3_object().content_length = self.model_version1.size
        s3_object().e_tag = f'"{self.model_version2.archive_hash}"'
        self.client.force_login(self.user1)
        request = self.build_model_version_update_request()
        with self.assertNumQueries(8):
            response = self.client.put(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}), request, format='json')

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": ['MD5 hashes do not match']})

    def test_create_model_requires_verified(self):
        """Model creation requires being verified
        """
        user = User.objects.create(display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        response = self.client.post(reverse('api:models'), {"name": "The Best Classifier"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_create_model_requires_logged_in(self):
        """Model creation requires being logged in
        """
        response = self.client.post(reverse('api:models'), {"name": "The Best Classifier"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_create_model_requires_name(self):
        """To create a model you must provide a name
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:models'))

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field is required."]})

    def test_create_model_requires_name_not_blank(self):
        """To create a model you must provide a name
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:models'), {"name": ''})

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field may not be blank."]})

    def test_create_model_new_name(self):
        """Create a model with a name no other model has and a description. The creator also gets admin rights on the newly created model.
        """
        self.client.force_login(self.user1)
        request = {
            "name": 'The Best Model Ever',
            "description": "This is actually the best model ever."
        }
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:models'), request)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()['rights'], ['read', 'write', 'admin'])

        model = Model.objects.get(id=response.json()['id'])
        self.assertEqual(model.name, request.get("name"))
        self.assertEqual(model.description, request.get("description"))

        # Check that user has admin rights on the model
        user_right = model.memberships.get(user=self.user1)
        self.assertEqual(user_right.level, Role.Admin.value)

    def test_create_model_name_taken_with_access(self):
        """Raises a 400 with the model_id when creating a model with a name that is already used for another model
        but with guest access rights on this model
        """
        self.client.force_login(self.user1)
        request = {
            "name": self.model1.name,
        }
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:models'), request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"id": str(self.model1.id), "name": 'A model with this name already exists'})

    def test_create_model_name_taken_no_access(self):
        """Raises a 403 with no additional information when creating a model with a name that is already used for another model
        but without access rights on this model
        """
        self.client.force_login(self.user3)
        request = {
            "name": self.model1.name,
        }
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:models'), request)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_list_models_requires_verified(self):
        user = User.objects.create(display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        response = self.client.get(reverse('api:models'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_list_models_requires_logged_in(self):
        response = self.client.get(reverse('api:models'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_list_models_low_access(self):
        """User 3 has only access to Model2 with a guest access.
        He only sees model_version3 of Model2 since it's the only one that
        has a set tag and is in Available state.
        """
        self.client.force_login(self.user3)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:models'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        models = response.json()['results']

        self.assertEqual(len(models), 1)
        self.assertListEqual(models, [_deserialize_model(self.model2, access_rights=['read'])])

    def test_list_models_contrib_access(self):
        """User 2 has contributor access to Model1 and Model2.
        He has contributor access on both that's why he sees all related versions.
        Models list is ordered by name, first Model1 (named 'First Model') then Model2 (named 'Second Model').
        """
        self.client.force_login(self.user2)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:models'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        models = response.json()['results']

        self.assertEqual(len(models), 2)
        self.assertListEqual(models, [
            _deserialize_model(self.model1, access_rights=['read', 'write']),
            _deserialize_model(self.model2, access_rights=['read', 'write', 'admin']),
        ])

    def test_list_models_filter_name(self):
        """User 2 has access to both Models, use search parameter name=Second returns only Model2
        """
        self.client.force_login(self.user2)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:models'), {'name': "second"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        models = response.json()['results']

        self.assertEqual(len(models), 1)
        self.assertListEqual(models, [_deserialize_model(self.model2, access_rights=['read', 'write', 'admin'])])

    def test_list_model_versions_requires_logged_in(self):
        """To list a model's versions, you need to be logged in.
        """
        response = self.client.get(reverse('api:model-versions', kwargs={"pk": str(self.model1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_list_model_versions_requires_guest_access(self):
        """To view a model's version, you need guest access on the model
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:model-versions', kwargs={"pk": str(self.model2.id)}))

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": 'You need a guest access to the model to list its versions.'})

    def test_list_model_versions_low_access(self):
        """With only guest access rights on a model, you only see the available versions with a set tag
        """
        self.client.force_login(self.user3)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:model-versions', kwargs={"pk": str(self.model2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual([version["id"] for version in response.json()['results']], [str(self.model_version3.id)])

    def test_list_model_versions_full_access(self):
        """With contributor access rights, you see every versions.
        """
        self.client.force_login(self.user2)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:model-versions', kwargs={"pk": str(self.model2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual([version["id"] for version in response.json()['results']], [str(self.model_version4.id), str(self.model_version3.id)])

    def test_destroy_model_versions_requires_admin(self):
        """To destroy a model version, you need admin rights on the model.
        """
        self.client.force_login(self.user2)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need an Admin access to the model to destroy this version."})

    def test_destroy_model_versions(self):
        """To destroy a model version, you need admin rights on the model.
        This also deletes every worker run that used this model version
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(10):
            response = self.client.delete(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version1.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_retrieve_model_versions_require_guest(self):
        """To retrieve a model version with a set tag and state==Available, you need guest rights on the model.
        """
        self.client.force_login(self.user1)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version3.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need a Guest access to the model to retrieve this version."})

    def test_retrieve_model_versions_tag_available(self):
        """Retrieve a model version with a set tag and state==Available with guest rights on the model.
        """
        self.client.force_login(self.user3)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version3.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), _deserialize_model_version(self.model_version3))

    def test_retrieve_model_versions_require_contributor(self):
        """To retrieve a model version with no set tag or state!=Available, you need contributor rights on the model.
        """
        self.client.force_login(self.user3)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version4.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the model to retrieve this version."})

    def test_retrieve_model_versions(self):
        """Retrieve a model version with no set tag or state!=Available with contributor rights on the model.
        """
        self.client.force_login(self.user2)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:model-version-retrieve', kwargs={"pk": str(self.model_version4.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), _deserialize_model_version(self.model_version4))

    def test_download_model_version_wrong_token(self):
        with self.assertNumQueries(1):
            response = self.client.get(reverse('api:model-version-download', kwargs={"pk": str(self.model_version1.id)}), {"token": "hehe_perhaps"})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_download_model_version_wrong_model_version_id(self):
        with self.assertNumQueries(1):
            response = self.client.get(reverse('api:model-version-download', kwargs={"pk": uuid4()}), {"token": "hehe_perhaps"})
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {'detail': 'Not found.'})

    def test_download_model_version(self):
        token = self.model_version1.build_authentication_token_hash()
        with self.assertNumQueries(1):
            response = self.client.get(reverse('api:model-version-download', kwargs={"pk": str(self.model_version1.id)}), {"token": token})
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.headers['Location'], self.model_version1.s3_url)
