import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import OAuthCredentials, OAuthStatus, User


class TestCredentials(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.cred1 = cls.user.credentials.get()
        cls.cred2 = cls.user.credentials.create(
            provider_name='gitlab',
            provider_url='https://somewhere.else',
            token='t0k3n',
            status=OAuthStatus.Error,
        )
        cls.user2 = User.objects.create(email='user2@test.test', display_name='User 2', verified_email=True)
        cls.cred3 = cls.user2.credentials.create(
            provider_name='gitlab',
            provider_url='https://somewhere.else',
            token='t0k3n',
            status=OAuthStatus.Created,
        )

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:credentials-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:credentials-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertCountEqual(data['results'], [
            {
                'id': str(self.cred1.id),
                'status': 'created',
                'provider_name': 'gitlab',
                'provider_display_name': 'GitLab',
                'provider_url': 'https://somewhere',
                'account_name': None,
            },
            {
                'id': str(self.cred2.id),
                'status': 'error',
                'provider_name': 'gitlab',
                'provider_display_name': 'GitLab',
                'provider_url': 'https://somewhere.else',
                'account_name': None,
            },
        ])

    def test_list_filtered(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:credentials-list'), {'status': 'error'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertCountEqual(data['results'], [
            {
                'id': str(self.cred2.id),
                'status': 'error',
                'provider_name': 'gitlab',
                'provider_display_name': 'GitLab',
                'provider_url': 'https://somewhere.else',
                'account_name': None,
            },
        ])

    def test_credentials_retrieve_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:credentials-retrieve', kwargs={'pk': str(self.cred1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_credentials_retrieve_not_found(self):
        fake_id = uuid.uuid4()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:credentials-retrieve', kwargs={'pk': str(fake_id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_credentials_retrieve(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:credentials-retrieve', kwargs={'pk': str(self.cred1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.cred1.id),
            'status': 'created',
            'provider_name': 'gitlab',
            'provider_display_name': 'GitLab',
            'provider_url': 'https://somewhere',
            'account_name': None,
        })

    def test_credentials_retrieve_wrong_user(self):
        """
        A user cannot retrieve another user's credentials
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:credentials-retrieve', kwargs={'pk': str(self.cred3.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_credentials_display_with_token(self):
        self.assertEqual(str(self.cred1), 'gitlab - user@user.fr - oauth-')

    def test_credentials_display_no_token(self):
        no_token = self.user.credentials.create(
            provider_name='gitlab',
            provider_url='https://somewhere.else',
            status=OAuthStatus.Created,
        )
        self.assertEqual(str(no_token), 'gitlab - user@user.fr')

    def test_destroy_credentials_requires_logged_in(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:credentials-retrieve', kwargs={"pk": str(self.cred2.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_destroy_credentials_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse('api:credentials-retrieve', kwargs={"pk": str(self.cred2.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_destroy_credentials(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:credentials-retrieve', kwargs={"pk": str(self.cred2.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        # Make sure it does not exist anymore
        with self.assertRaises(OAuthCredentials.DoesNotExist):
            self.cred2.refresh_from_db()

    def test_destroy_credentials_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse('api:credentials-retrieve', kwargs={"pk": str(uuid.uuid4())}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
