import urllib.parse
from unittest.mock import MagicMock, patch

import responses
from django.http.request import HttpRequest
from django.test import override_settings
from requests import HTTPError
from responses import matchers

from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import OAuthStatus
from arkindex.users.providers import GitLabOAuthProvider


class TestGitLabOAuthProvider(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.gl_patch = patch('arkindex.users.providers.Gitlab')

    def setUp(self):
        super().setUp()
        self.gl_mock = self.gl_patch.start()

    def tearDown(self):
        self.gl_patch.stop()
        super().tearDown()

    def test_enabled(self):
        with self.settings(GITLAB_APP_ID=None, GITLAB_APP_SECRET=None):
            self.assertFalse(GitLabOAuthProvider.enabled())
        with self.settings(GITLAB_APP_ID='abcd', GITLAB_APP_SECRET='1234'):
            self.assertTrue(GitLabOAuthProvider.enabled())

    @override_settings(BACKEND_PUBLIC_URL_OAUTH='http://arkindex.localhost:8000', PUBLIC_HOSTNAME='http://arkindex.localhost:8080')
    def test_callback_uri_dev_override(self):
        self.assertEqual(GitLabOAuthProvider().get_callback_uri(), 'http://arkindex.localhost:8000/api/v1/oauth/providers/gitlab/callback/')

    @override_settings(BACKEND_PUBLIC_URL_OAUTH=None, PUBLIC_HOSTNAME='https://arkindex.localhost/')
    def test_callback_uri_public_hostname(self):
        self.assertEqual(GitLabOAuthProvider().get_callback_uri(), 'https://arkindex.localhost/api/v1/oauth/providers/gitlab/callback/')

    @override_settings(BACKEND_PUBLIC_URL_OAUTH=None, PUBLIC_HOSTNAME=None)
    def test_callback_uri_request(self):
        with self.assertRaisesMessage(AssertionError, 'PUBLIC_HOSTNAME is required to generate callback URIs'):
            GitLabOAuthProvider().get_callback_uri()

    @override_settings(BACKEND_PUBLIC_URL_OAUTH=None, PUBLIC_HOSTNAME='https://arkindex.localhost/')
    def test_authorize_uri(self):
        with self.settings(GITLAB_APP_ID='abcd'):
            uri = GitLabOAuthProvider(credentials=self.creds).get_authorize_uri()
        parsed = urllib.parse.urlparse(uri)
        self.assertEqual(parsed.scheme, 'https')
        self.assertEqual(parsed.netloc, 'somewhere')
        self.assertEqual(parsed.path, '/oauth/authorize')
        query = urllib.parse.parse_qs(parsed.query)
        self.assertDictEqual(query, {
            'client_id': ['abcd'],
            'redirect_uri': ['https://arkindex.localhost/api/v1/oauth/providers/gitlab/callback/'],
            'scope': ['api'],
            'response_type': ['code'],
            'state': [str(self.creds.id)],
        })

    def test_handle_callback_no_state(self):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {}
        with self.assertRaisesMessage(ValueError, 'No state hash'):
            GitLabOAuthProvider().handle_callback(request_mock)

    def test_handle_callback_bad_response(self):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {'state': str(self.creds.id)}
        with self.assertRaisesRegex(ValueError, 'valid response'):
            GitLabOAuthProvider().handle_callback(request_mock)

    def test_handle_callback_error(self):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {'state': str(self.creds.id), 'error': 'error message'}
        with self.assertRaisesMessage(ValueError, 'error message'):
            GitLabOAuthProvider().handle_callback(request_mock)

    def test_handle_callback_overwrite(self):
        self.creds.status = OAuthStatus.Done
        self.creds.save()
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {'state': str(self.creds.id), 'code': 'something'}
        with self.assertRaisesRegex(Exception, 'overwrite'):
            GitLabOAuthProvider().handle_callback(request_mock)

    @responses.activate
    @override_settings(BACKEND_PUBLIC_URL_OAUTH=None, PUBLIC_HOSTNAME='https://arkindex.localhost')
    def test_handle_callback_success(self):
        responses.post(
            'https://somewhere/oauth/token',
            match=[
                matchers.urlencoded_params_matcher({
                    'client_id': 'abcd',
                    'client_secret': 's3kr3t',
                    'code': 'abc123',
                    'redirect_uri': 'https://arkindex.localhost/api/v1/oauth/providers/gitlab/callback/',
                    'grant_type': 'authorization_code',
                }),
            ],
            json={
                'access_token': 't0k3n',
                'refresh_token': 'r3fr3sh',
                'token_type': 'Bearer',
                'created_at': 1582984800,
                'expires_in': '3600',
            },
        )
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {'state': str(self.creds.id), 'code': 'abc123'}
        request_mock.build_absolute_uri.return_value = 'callback'

        self.gl_mock.return_value.user.username = 'bobby'

        with self.settings(GITLAB_APP_ID='abcd', GITLAB_APP_SECRET='s3kr3t'):
            GitLabOAuthProvider().handle_callback(request_mock)

        self.creds.refresh_from_db()
        self.assertEqual(self.creds.token, 't0k3n')
        self.assertEqual(self.creds.refresh_token, 'r3fr3sh')
        self.assertEqual(self.creds.expiry.isoformat(), '2020-02-29T15:00:00+00:00')
        self.assertEqual(self.creds.account_name, 'bobby')
        self.assertEqual(self.creds.status, OAuthStatus.Done)
        self.assertEqual(self.gl_mock.return_value.auth.call_count, 1)

        args, kwargs = self.gl_mock.call_args
        self.assertEqual(args, ('https://somewhere', ))
        self.assertDictEqual(kwargs, {'oauth_token': 't0k3n'})

    @responses.activate
    @override_settings(BACKEND_PUBLIC_URL_OAUTH=None, PUBLIC_HOSTNAME='https://arkindex.localhost')
    def test_refresh_token(self):
        responses.post(
            'https://somewhere/oauth/token',
            match=[
                matchers.urlencoded_params_matcher({
                    'client_id': 'abcd',
                    'client_secret': 's3kr3t',
                    'redirect_uri': 'https://arkindex.localhost/api/v1/oauth/providers/gitlab/callback/',
                    'grant_type': 'refresh_token',
                    'refresh_token': 'r3fr3sh'
                }),
            ],
            json={
                'access_token': 't0k3n',
                'refresh_token': 'r4fr4sh',
                'token_type': 'Bearer',
                'created_at': 1582984800,
                'expires_in': '3600',
            },
        )

        self.gl_mock.return_value.user.username = 'bobby'

        self.creds.refresh_token = 'r3fr3sh'
        self.creds.save()
        self.assertEqual(self.creds.expiry.isoformat(), '2100-12-31T23:59:59.999000+00:00')
        self.assertEqual(self.creds.token, 'oauth-token')

        with self.settings(GITLAB_APP_ID='abcd', GITLAB_APP_SECRET='s3kr3t'):
            GitLabOAuthProvider(credentials=self.creds).refresh_token()

        self.assertEqual(self.creds.token, 't0k3n')
        self.assertEqual(self.creds.refresh_token, 'r4fr4sh')
        self.assertEqual(self.creds.expiry.isoformat(), '2020-02-29T15:00:00+00:00')
        self.assertEqual(self.creds.account_name, 'bobby')
        self.assertEqual(self.creds.status, OAuthStatus.Done)
        self.assertEqual(self.gl_mock.return_value.auth.call_count, 1)

        args, kwargs = self.gl_mock.call_args
        self.assertEqual(args, ('https://somewhere', ))
        self.assertDictEqual(kwargs, {'oauth_token': 't0k3n'})

    @responses.activate
    @override_settings(BACKEND_PUBLIC_URL_OAUTH=None, PUBLIC_HOSTNAME='https://arkindex.localhost')
    def test_refresh_token_error(self):
        responses.post(
            'https://somewhere/oauth/token',
            match=[
                matchers.urlencoded_params_matcher({
                    'client_id': 'abcd',
                    'client_secret': 's3kr3t',
                    'redirect_uri': 'https://arkindex.localhost/api/v1/oauth/providers/gitlab/callback/',
                    'grant_type': 'refresh_token',
                    'refresh_token': 'r3fr3sh'
                }),
            ],
            status=418,
        )

        self.creds.refresh_token = 'r3fr3sh'
        self.creds.save()
        self.assertEqual(self.creds.expiry.isoformat(), '2100-12-31T23:59:59.999000+00:00')
        self.assertEqual(self.creds.token, 'oauth-token')

        with self.settings(GITLAB_APP_ID='abcd', GITLAB_APP_SECRET='s3kr3t'), self.assertRaises(HTTPError):
            GitLabOAuthProvider(credentials=self.creds).refresh_token()

        self.assertEqual(self.creds.token, 'oauth-token')
        self.assertEqual(self.creds.refresh_token, 'r3fr3sh')
        self.assertEqual(self.creds.expiry.isoformat(), '2100-12-31T23:59:59.999000+00:00')
        self.assertEqual(self.creds.status, OAuthStatus.Error)
        self.assertFalse(self.gl_mock.return_value.auth.called)

    @override_settings(BACKEND_PUBLIC_URL_OAUTH=None, PUBLIC_HOSTNAME='https://arkindex.localhost')
    def test_disconnect(self):
        repo_1, repo_2 = self.creds.repos.order_by('hook_token')
        # This hook does not match the expected hook URL and should not be removed
        ignored_hook = MagicMock()
        ignored_hook.url = 'https://potato.localhost/something'
        repo_1_hook = MagicMock()
        repo_1_hook.url = f'https://arkindex.localhost/api/v1/imports/hook/{repo_1.id}/'
        repo_2_hook = MagicMock()
        repo_2_hook.url = f'https://arkindex.localhost/api/v1/imports/hook/{repo_2.id}/'
        self.gl_mock().projects.get.return_value.id = 'repo_id'
        self.gl_mock().projects.get.return_value.hooks.list.return_value = [
            ignored_hook, repo_1_hook, repo_2_hook
        ]

        GitLabOAuthProvider(credentials=self.creds).disconnect()

        self.assertEqual(self.gl_mock().projects.get.call_count, 2)
        self.assertEqual(self.gl_mock().projects.get.return_value.hooks.list.call_count, 2)
        self.assertFalse(ignored_hook.delete.called)
        self.assertEqual(repo_1_hook.delete.call_count, 1)
        self.assertEqual(repo_2_hook.delete.call_count, 1)
